# -*- coding: utf-8 -*-

# Copyright (c) 2016, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

"""Rename ServerOGC to OGCServer

Revision ID: 6a412d9437b1
Revises: 29f2a32859ec
Create Date: 2016-06-28 18:08:23.888198
"""

from alembic import op, context

# revision identifiers, used by Alembic.
revision = '6a412d9437b1'
down_revision = '29f2a32859ec'
branch_labels = None
depends_on = None


def upgrade():
    schema = context.get_context().config.get_main_option('schema')

    op.rename_table('server_ogc', 'ogc_server', schema=schema)
    with op.batch_alter_table('layer_wms', schema=schema) as table_op:
        table_op.alter_column('server_ogc_id', new_column_name='ogc_server_id')


def downgrade():
    schema = context.get_context().config.get_main_option('schema')

    op.rename_table('ogc_server', 'server_ogc', schema=schema)
    with op.batch_alter_table('layer_wms', schema=schema) as table_op:
        table_op.alter_column('ogc_server_id', new_column_name='server_ogc_id')



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2016, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import argparse
import transaction
from pyramid.paster import get_app


def main():
    parser = argparse.ArgumentParser(
        description="Create and populate the database tables."
    )
    parser.add_argument(
        '-i', '--iniconfig',
        default='production.ini',
        help='project .ini config file'
    )
    parser.add_argument(
        '-n', '--app-name',
        default="app",
        help='The application name (optional, default is "app")'
    )

    options = parser.parse_args()

    # read the configuration
    get_app(options.iniconfig, options.app_name)

    from c2cgeoportal.models import DBSession, Interface, OGCServer, Theme, LayerGroup, LayerWMS

    session = DBSession()

    interfaces = session.query(Interface).all()
    ogc_server = session.query(OGCServer).filter(OGCServer.name == u"source for image/png").one()

    layer_borders = LayerWMS(u"Borders", u"borders")
    layer_borders.interfaces = interfaces
    layer_borders.ogc_server = ogc_server
    layer_density = LayerWMS(u"Density", u"density")
    layer_density.interfaces = interfaces
    layer_density.ogc_server = ogc_server

    group = LayerGroup(u"Demo")
    group.children = [layer_borders, layer_density]

    theme = Theme(u"Demo")
    theme.children = [group]
    theme.interfaces = interfaces

    transaction.commit()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Copyright (c) 2012-2016, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import math
import geojson
from decimal import Decimal

from pyramid.view import view_config
from pyramid.response import Response
from pyramid.httpexceptions import HTTPNotFound
from pyramid.i18n import TranslationStringFactory

from c2cgeoportal.views.raster import Raster
from c2cgeoportal.lib.caching import set_common_headers, NO_CACHE


_ = TranslationStringFactory("c2cgeoportal")


class Profile(Raster):

    def __init__(self, request):
        self.request = request
        Raster.__init__(self, request)

    @view_config(route_name="profile.json", renderer="decimaljson")
    def json(self):
        """answers to /profile.json"""
        layer, points = self._compute_points()
        set_common_headers(
            self.request, "profile", NO_CACHE
        )
        return {"profile": points}

    @view_config(route_name="profile.csv")
    def csv(self):
        """answers to /profile.csv"""
        layers, points = self._compute_points()

        result = _("distance") + "," + ",".join(layers) + ",x,y"
        template = ",".join("%s" for l in layers)
        for point in points:
            # Handles cases when a layer is undefined, thus when not all raster
            # have the same geographical coverage
            for l in layers:
                if l not in point["values"]:
                    point["values"][l] = -9999

            r = template % tuple((str(point["values"][l]) for l in layers))
            result += "\n{},{},{:.1f},{:.1f}".format(str(point["dist"]), r, point["x"], point["y"])

        return set_common_headers(
            self.request, "profile", NO_CACHE,
            response=Response(result, headers={
                "Content-Disposition": 'attachment; filename="profile.csv"',
            }), content_type="text/csv; charset=utf-8", vary=True
        )

    def _compute_points(self):
        """Compute the alt=fct(dist) array"""
        geom = geojson.loads(self.request.params["geom"], object_hook=geojson.GeoJSON.to_instance)

        if "layers" in self.request.params:
            rasters = {}
            layers = self.request.params["layers"].split(",")
            for layer in layers:
                if layer in self.rasters:
                    rasters[layer] = self.rasters[layer]
                else:
                    raise HTTPNotFound("Layer {0!s} not found".format(layer))
        else:
            rasters = self.rasters

        points = []

        dist = 0
        prev_coord = None
        coords = self._create_points(geom.coordinates, int(self.request.params["nbPoints"]))
        for coord in coords:
            if prev_coord is not None:
                dist += self._dist(prev_coord, coord)

            values = {}
            has_one = False
            for ref in rasters.keys():
                value = self._get_raster_value(
                    self.rasters[ref],
                    ref, coord[0], coord[1])
                if value is not None:
                    values[ref] = value
                    has_one = True

            if has_one:
                # 10cm accuracy is enough for distances
                rounded_dist = Decimal(str(dist)).quantize(Decimal("0.1"))
                points.append({
                    "dist": rounded_dist,
                    "values": values,
                    "x": coord[0],
                    "y": coord[1]
                })
            prev_coord = coord

        return rasters.keys(), points

    @staticmethod
    def _dist(coord1, coord2):
        """Compute the distance between 2 points"""
        return math.sqrt(math.pow(coord1[0] - coord2[0], 2.0) +
                         math.pow(coord1[1] - coord2[1], 2.0))

    def _create_points(self, coords, nb_points):
        """Add some points in order to reach roughly the asked number of points"""
        total_length = 0
        prev_coord = None
        for coord in coords:
            if prev_coord is not None:
                total_length += self._dist(prev_coord, coord)
            prev_coord = coord

        if total_length == 0.0:
            return coords

        result = []
        prev_coord = None
        for coord in coords:
            if prev_coord is not None:
                cur_length = self._dist(prev_coord, coord)
                cur_nb_points = int(nb_points * cur_length / total_length + 0.5)
                if cur_nb_points < 1:
                    cur_nb_points = 1
                dx = (coord[0] - prev_coord[0]) / float(cur_nb_points)
                dy = (coord[1] - prev_coord[1]) / float(cur_nb_points)
                for i in range(1, cur_nb_points + 1):
                    result.append([prev_coord[0] + dx * i, prev_coord[1] + dy * i])
            else:
                result.append([coord[0], coord[1]])
            prev_coord = coord
        return result



# Copyright (c) 2016, Camptocamp SA
# All rights reserved.

# Auto-generated file. Do not Edit!
INFO = {
    "git_hash": "54cb1f3b9479f36805596df042d8f63286d76a54",
    "version": "2.1",
    "git_tag": "2.1.0.dev23",
}



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2016, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import os
from mako.template import Template
from mako.lookup import TemplateLookup


def main():
    """ This function build the print configuration file config.yaml using
    mako.
    The file is created in the folder print/.
    If the file print/templates/print.mako doesnt exists, the build is not
    executed
    """

    base_template = 'print/templates/print.mako'
    if os.path.exists(base_template):
        print "building print template"

        # DONT! add a trailing / to the lookup path, mako cant find templates otherwise
        mylookup = TemplateLookup(directories=['print/templates'])
        mytemplate = Template(filename=base_template, lookup=mylookup)

        print_template = open('print/config.yaml', 'w+')
        print_template.write(mytemplate.render().encode('UTF-8'))
        print_template.close()

        print "finished building print template"
    else:
        print "no print template found, ignoring"



# -*- coding: utf-8 -*-

# Copyright (c) 2015-2016, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import logging

from pyramid.view import view_config
from xml.etree import ElementTree
from pyramid.httpexceptions import HTTPForbidden, HTTPBadRequest, HTTPUnauthorized

from c2cgeoportal.lib.caching import get_region, NO_CACHE, PRIVATE_CACHE
from c2cgeoportal.lib.filter_capabilities import filter_wfst_capabilities, \
    normalize_tag, normalize_typename, get_writable_layers
from c2cgeoportal.views.proxy import Proxy

cache_region = get_region()
log = logging.getLogger(__name__)


class TinyOWSProxy(Proxy):

    def __init__(self, request):
        Proxy.__init__(self, request)
        self.settings = request.registry.settings.get("tinyowsproxy", {})

        assert \
            "tinyows_url" in self.settings, \
            "tinyowsproxy.tinyows_url must be set"

    def _get_wfs_url(self):
        return self.settings.get("tinyows_url")

    @view_config(route_name="tinyowsproxy")
    def proxy(self):
        self.user = self.request.user

        if self.user is None:
            raise HTTPUnauthorized(
                "Authentication required",
                headers=[("WWW-Authenticate", 'Basic realm="TinyOWS"')]
            )
        self.role_id = None if self.user is None else self.user.role.id

        # params hold the parameters we're going to send to TinyOWS
        params = dict(self.request.params)
        self.lower_params = self._get_lower_params(params)

        operation = self.lower_params.get("request")
        typenames = \
            set([normalize_typename(self.lower_params.get("typename"))]) \
            if "typename" in self.lower_params \
            else set()

        method = self.request.method
        if method == "POST":
            try:
                (operation, typenames_post) = \
                    self._parse_body(self.request.body)
            except Exception as e:
                log.error("Error while parsing POST request body")
                log.exception(e)
                raise HTTPBadRequest(
                    "Error parsing the request (see logs for more details)")

            typenames = typenames.union(typenames_post)

        if operation is None or operation == "":
            operation = "getcapabilities"

        if operation == "describefeaturetype":
            # for DescribeFeatureType we require that exactly one type-name
            # is given, otherwise we would have to filter the result
            if len(typenames) != 1:
                raise HTTPBadRequest(
                    "Exactly one type-name must be given for "
                    "DescribeFeatureType requests"
                )

        if not self._is_allowed(typenames):
            raise HTTPForbidden(
                "No access rights for at least one of the given type-names"
            )

        # we want clients to cache GetCapabilities and DescribeFeatureType req.
        use_cache = method == "GET" and operation in (u"getcapabilities", u"describefeaturetype")
        cache_control = PRIVATE_CACHE if use_cache else NO_CACHE

        response = self._proxy_callback(
            operation, self.role_id, cache_control,
            url=self._get_wfs_url(), params=params, cache=use_cache,
            headers=self._get_headers(), body=self.request.body,
        )
        return response

    def _is_allowed(self, typenames):
        """
        Checks if the current user has the rights to access the given
        type-names.
        """
        writable_layers = get_writable_layers(self.role_id)
        return typenames.issubset(writable_layers)

    def _get_headers(self):
        headers = Proxy._get_headers(self)
        if "tinyows_host" in self.settings:
            headers["Host"] = self.settings.get("tinyows_host")
        return headers

    def _proxy_callback(self, operation, role_id, cache_control, *args, **kwargs):
        resp, content = self._proxy(*args, **kwargs)

        if operation == "getcapabilities":
            content = filter_wfst_capabilities(
                content, role_id,
                self._get_wfs_url(),
                self.settings.get("proxies")
            )

        content = self._filter_urls(
            content,
            self.settings.get("online_resource"),
            self.settings.get("proxy_online_resource")
        )

        return self._build_response(
            resp, content, cache_control, "tinyows"
        )

    @staticmethod
    def _filter_urls(content, online_resource, proxy_online_resource):
        if online_resource is not None and proxy_online_resource is not None:
            return content.replace(online_resource, proxy_online_resource)
        else:
            return content

    @staticmethod
    def _parse_body(body):
        """
        Read the WFS-T request body and extract the referenced type-names
        and request method.
        """
        xml = ElementTree.fromstring(body)
        wfs_request = normalize_tag(xml.tag)

        # get the type names
        typenames = set()
        for child in xml:
            tag = normalize_tag(child.tag)
            if tag == "typename":
                typenames.add(child.text)
            elif tag in ("query", "lock", "update", "delete"):
                typenames.add(child.get("typeName"))
            elif tag == "insert":
                for insert_child in child:
                    typenames.add(normalize_tag(insert_child.tag))

        # remove the namespace from the typenames
        typenames = {normalize_typename(t) for t in typenames}

        return (wfs_request, typenames)



# -*- coding: utf-8 -*-

# Copyright (c) 2014-2016, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import sys
import json
import argparse
import subprocess


def main():
    parser = argparse.ArgumentParser(
        prog=sys.argv[0], add_help=True,
        description="Tool used to migrate the localisation v1 to v2.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""The source file should be wheel formated as:
OpenLayers.Util.extend(OpenLayers.Lang.<lang>, {
    ...
});

Carful, It will truncate the destination file!

After running this script you can do:
touch <package>/locale/<package>-client.pot
and build your application to merge the old localisation with the new one.
""",
    )

    parser.add_argument(
        "lang",
        help="the language to translate"
    )
    parser.add_argument(
        "json_v1",
        help="the JSON l10n file from the version 1"
    )
    parser.add_argument(
        "po_v2",
        help="the po file for the version 2"
    )
    options = parser.parse_args()

    with open(options.json_v1) as src:
        lines = src.readlines()
        while lines[-1].strip() == "":
            lines = lines[0:-1]

        jsonlines = ["console.log(JSON.stringify({"]
        jsonlines += lines[1:-1]
        jsonlines.append("}))")

        json_ = subprocess.check_output(["node", "-e", " ".join(jsonlines)])
        source = json.loads(json_)

    with open(options.po_v2, "w+") as destionation:
        destionation.write("""msgid ""
msgstr ""
"Last-Translator: Imported from {0!s}\\n"
"Language: {1!s}\\n"
"MIME-Version: 1.0\\n"
"Content-Type: text/plain; charset=UTF-8\\n"
"Content-Transfer-Encoding: 8bit\\n"
"Plural-Forms: nplurals=2; plural=(n != 1);\\n"
""".format(options.json_v1, options.lang))
        for key, value in source.items():
            if isinstance(value, basestring):
                destionation.write(("""
msgid "{0!s}"
msgstr "{1!s}"
""".format(key, value.replace('"', '\\"'))).encode("utf-8"),
                )



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2016, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import logging

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPBadRequest
from c2cgeoportal.views.proxy import Proxy
from c2cgeoportal.lib.caching import NO_CACHE

import ast

log = logging.getLogger(__name__)


class ResourceProxy(Proxy):

    def __init__(self, request):  # pragma: no cover
        Proxy.__init__(self, request)
        self.request = request
        self.settings = request.registry.settings.get("resourceproxy", {})

    @view_config(route_name="resourceproxy")
    def proxy(self):  # pragma: no cover
        target = self.request.params.get("target", "")
        targets = self.settings.get("targets", [])
        if target in targets:  # pragma: no cover
            url = targets[target]
            values = ast.literal_eval(self.request.params.get("values"))
            url = url % values

            resp, content = self._proxy(url=url)

            cache_control = NO_CACHE
            content_type = resp["content-type"]

            return self._build_response(
                resp, content, cache_control, "externalresource",
                content_type=content_type
            )
        else:  # pragma: no cover
            log.warning("target url not found: {0!s}".format(target))
            return HTTPBadRequest("url not allowed")



# -*- coding: utf-8 -*-

# Copyright (c) 2012-2016, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


# file taken from http://indiemaps.com/blog/2008/03/easy-shapefile-loading-in-python/

import math

from struct import unpack
import dbfutils


XY_POINT_RECORD_LENGTH = 16
db = []


def load_shapefile(file_name):
    global db
    file_name = file_name
    records = []
    # open dbf file and get records as a list
    dbf_file = file_name[0:-4] + ".dbf"
    dbf = open(dbf_file, "rb")
    db = list(dbfutils.dbfreader(dbf))
    dbf.close()
    fp = open(file_name, "rb")

    # get basic shapefile configuration
    fp.seek(32)
    read_and_unpack("i", fp.read(4))
    read_bounding_box(fp)

    # fetch Records
    fp.seek(100)
    while True:
        shp_record = create_record(fp)
        if not shp_record:
            break
        records.append(shp_record)

    return records


record_class = {
    0: "RecordNull",
    1: "RecordPoint",
    8: "RecordMultiPoint",
    3: "RecordPolyLine",
    5: "RecordPolygon"
}


def create_record(fp):
    # read header
    record_number = read_and_unpack(">L", fp.read(4))
    if record_number == "":
        return False
    read_and_unpack(">L", fp.read(4))
    record_shape_type = read_and_unpack("<L", fp.read(4))

    shp_data = read_record_any(fp, record_shape_type)
    dbf_data = {}
    for i in range(0, len(db[record_number + 1])):
        dbf_data[db[0][i]] = db[record_number + 1][i]

    return {"shp_data": shp_data, "dbf_data": dbf_data}


# Reading defs


def read_record_any(fp, type):
    if type == 0:
        return read_record_null(fp)
    elif type == 1:
        return read_record_point(fp)
    elif type == 8:
        return read_record_multi_point(fp)
    elif type in [3, 5]:
        return read_record_poly_line(fp)
    else:
        return False


def read_record_null(fp):
    data = {}
    return data


point_count = 0


def read_record_point(fp):
    global point_count
    data = {}
    data["x"] = read_and_unpack("d", fp.read(8))
    data["y"] = read_and_unpack("d", fp.read(8))
    point_count += 1
    return data


def read_record_multi_point(fp):
    data = read_bounding_box(fp)
    data["numpoints"] = read_and_unpack("i", fp.read(4))
    for i in range(0, data["numpoints"]):
        data["points"].append(read_record_point(fp))
    return data


def read_record_poly_line(fp):
    data = read_bounding_box(fp)
    data["numparts"] = read_and_unpack("i", fp.read(4))
    data["numpoints"] = read_and_unpack("i", fp.read(4))
    data["parts"] = []
    for i in range(0, data["numparts"]):
        data["parts"].append(read_and_unpack("i", fp.read(4)))
    points_initial_index = fp.tell()
    points_read = 0
    for part_index in range(0, data["numparts"]):
        # if(!isset(data["parts"][part_index]["points"]) or
        # !is_array(data["parts"][part_index]["points"])):
        data["parts"][part_index] = {}
        data["parts"][part_index]["points"] = []

        # while( ! in_array( points_read, data["parts"]) and
        # points_read < data["numpoints"] and !feof(fp)):
        check_point = []
        while (points_read < data["numpoints"]):
            curr_point = read_record_point(fp)
            data["parts"][part_index]["points"].append(curr_point)
            points_read += 1
            if points_read == 0 or check_point == []:
                check_point = curr_point
            elif curr_point == check_point:
                check_point = []
                break

    fp.seek(points_initial_index + (points_read * XY_POINT_RECORD_LENGTH))
    return data


# General defs


def read_bounding_box(fp):
    data = {}
    data["xmin"] = read_and_unpack("d", fp.read(8))
    data["ymin"] = read_and_unpack("d", fp.read(8))
    data["xmax"] = read_and_unpack("d", fp.read(8))
    data["ymax"] = read_and_unpack("d", fp.read(8))
    return data


def read_and_unpack(type, data):
    if data == "":
        return data
    return unpack(type, data)[0]


# ###
# ### additional functions
# ###


def get_centroids(records, projected=False):
    # for each feature
    if projected:
        points = "projectedPoints"
    else:
        points = "points"

    for feature in records:
        numpoints = cx = cy = 0
        for part in feature["shp_data"]["parts"]:
            for point in part[points]:
                numpoints += 1
                cx += point["x"]
                cy += point["y"]
        cx /= numpoints
        cy /= numpoints
        feature["shp_data"]["centroid"] = {"x": cx, "y": cy}


def get_bound_centers(records):
    for feature in records:
        cx = .5 * (feature["shp_data"]["xmax"] - feature["shp_data"]["xmin"]) + \
            feature["shp_data"]["xmin"]
        cy = .5 * (feature["shp_data"]["ymax"] - feature["shp_data"]["ymin"]) + \
            feature["shp_data"]["ymin"]
        feature["shp_data"]["boundCenter"] = {"x": cx, "y": cy}


def get_true_centers(records, projected=False):
    # gets the true polygonal centroid for each feature (uses largest ring)
    # should be spherical, but isn't

    if projected:
        points = "projectedPoints"
    else:
        points = "points"

    for feature in records:
        maxarea = 0
        for ring in feature["shp_data"]["parts"]:
            ring_area = get_area(ring, points)
            if ring_area > maxarea:
                maxarea = ring_area
                biggest = ring
            # now get the true centroid
        temp_point = {"x": 0, "y": 0}
        if biggest[points][0] != biggest[points][len(biggest[points]) - 1]:
            print  \
                "mug", biggest[points][0], \
                biggest[points][len(biggest[points]) - 1]
        for i in range(0, len(biggest[points]) - 1):
            j = (i + 1) % (len(biggest[points]) - 1)
            temp_point["x"] -= \
                (biggest[points][i]["x"] + biggest[points][j]["x"]) * \
                (
                    (biggest[points][i]["x"] * biggest[points][j]["y"]) -
                    (biggest[points][j]["x"] * biggest[points][i]["y"]))
            temp_point["y"] -= \
                (biggest[points][i]["y"] + biggest[points][j]["y"]) * \
                (
                    (biggest[points][i]["x"] * biggest[points][j]["y"]) -
                    (biggest[points][j]["x"] * biggest[points][i]["y"]))

        temp_point["x"] = temp_point["x"] / ((6) * maxarea)
        temp_point["y"] = temp_point["y"] / ((6) * maxarea)
        feature["shp_data"]["truecentroid"] = temp_point


def get_area(ring, points):
    # returns the area of a polygon
    # needs to be spherical area, but isn"t
    area = 0
    for i in range(0, len(ring[points]) - 1):
        j = (i + 1) % (len(ring[points]) - 1)
        area += ring[points][i]["x"] * ring[points][j]["y"]
        area -= ring[points][i]["y"] * ring[points][j]["x"]

    return math.fabs(area / 2)


def get_neighbors(records):
    # for each feature
    for i in range(len(records)):
        if "neighbors" not in records[i]["shp_data"]:
            records[i]["shp_data"]["neighbors"] = []

        # for each other feature
        for j in range(i + 1, len(records)):
            numcommon = 0
            # first check to see if the bounding boxes overlap
            if overlap(records[i], records[j]):
                # if so, check every single point in this feature
                # to see if it matches a point in the other feature

                # for each part:
                for part in records[i]["shp_data"]["parts"]:

                    # for each point:
                    for point in part["points"]:

                        for otherPart in records[j]["shp_data"]["parts"]:
                            if point in otherPart["points"]:
                                numcommon += 1
                                if numcommon == 2:
                                    if "neighbors" not in records[j]["shp_data"]:
                                        records[j]["shp_data"]["neighbors"] = []
                                    records[i]["shp_data"]["neighbors"].append(j)
                                    records[j]["shp_data"]["neighbors"].append(i)
                                    # now break out to the next j
                                    break
                        if numcommon == 2:
                            break
                    if numcommon == 2:
                        break


def project_shapefile(records, what_projection, lon_center=0, lat_center=0):
    print "projecting to ", what_projection
    for feature in records:
        for part in feature["shp_data"]["parts"]:
            part["projectedPoints"] = []
            for point in part["points"]:
                temp_point = project_point(
                    point, what_projection, lon_center, lat_center)
                part["projectedPoints"].append(temp_point)


def project_point(from_point, what_projection, lon_center, lat_center):
    lat_radians = from_point["y"] * math.pi / 180
    if lat_radians > 1.5:
        lat_radians = 1.5
    if lat_radians < -1.5:
        lat_radians = -1.5
    lon_radians = from_point["x"] * math.pi / 180
    lon_center = lon_center * math.pi / 180
    lat_center = lat_center * math.pi / 180
    new_point = {}
    if what_projection == "MERCATOR":
        new_point["x"] = (180 / math.pi) * (lon_radians - lon_center)
        new_point["y"] = (180 / math.pi) * math.log(
            math.tan(lat_radians) + (1 / math.cos(lat_radians)))
        if new_point["y"] > 200:
            new_point["y"] = 200
        if new_point["y"] < -200:
            new_point["y"] = 200
        return new_point
    if what_projection == "EQUALAREA":
        new_point["x"] = 0
        new_point["y"] = 0
        return new_point


def overlap(feature1, feature2):
    if (feature1["shp_data"]["xmax"] > feature2["shp_data"]["xmin"] and
            feature1["shp_data"]["ymax"] > feature2["shp_data"]["ymin"] and
            feature1["shp_data"]["xmin"] < feature2["shp_data"]["xmax"] and
            feature1["shp_data"]["ymin"] < feature2["shp_data"]["ymax"]):
        return True
    else:
        return False



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2016, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


from pyramid.security import Allow, Authenticated, ALL_PERMISSIONS
from pyramid_formalchemy.resources import Models


class Root:
    def __init__(self, request):
        self.request = request  # pragma: no cover


class FAModels(Models):
    __acl__ = [
        (Allow, Authenticated, ALL_PERMISSIONS),
    ]


def defaultgroupsfinder(username, request):
    """ The c2cgeoportal default group finder. To be used as the callback of
    the ``AuthTktAuthenticationPolicy`` or any callback-based authentication
    policy. """
    if not hasattr(request, "user") or request.user is None:
        return []  # pragma: no cover
    role = request.user.role
    return [role.name] if role else []



# -*- coding: utf-8 -*-

# Copyright (c) 2012-2016, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import logging
from decimal import Decimal

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPInternalServerError, HTTPNotFound

from c2cgeoportal.lib.raster.georaster import GeoRaster
from c2cgeoportal.lib.caching import set_common_headers, NO_CACHE

log = logging.getLogger(__name__)


class Raster:

    # cache of GeoRaster instances in function of the layer name
    _rasters = {}

    def __init__(self, request):
        self.request = request
        self.rasters = self.request.registry.settings["raster"]

    @view_config(route_name="raster", renderer="decimaljson")
    def raster(self):
        lon = float(self.request.params["lon"])
        lat = float(self.request.params["lat"])
        if "layers" in self.request.params:
            rasters = {}
            layers = self.request.params["layers"].split(",")
            for layer in layers:
                if layer in self.rasters:
                    rasters[layer] = self.rasters[layer]
                else:
                    raise HTTPNotFound("Layer {0!s} not found".format(layer))
        else:
            rasters = self.rasters

        result = {}
        for ref in rasters.keys():
            result[ref] = self._get_raster_value(
                rasters[ref], ref, lon, lat)

        set_common_headers(
            self.request, "raster", NO_CACHE
        )
        return result

    def _get_raster_value(self, layer, ref, lon, lat):
        if ref in self._rasters:
            raster = self._rasters[ref]
        elif "type" not in layer or layer["type"] == "shp_index":
            raster = GeoRaster(layer["file"])
            self._rasters[ref] = raster
        else:  # pragma: no cover
            raise HTTPInternalServerError(
                'Bad raster type "{0!s}" for raster layer "{1!s}"'.format(
                    layer["type"], ref
                )
            )

        result = raster.get_value(lon, lat)
        if "round" in layer:
            result = self._round(result, layer["round"])
        elif result is not None:
            result = Decimal(str(result))

        return result

    @staticmethod
    def _round(value, round_to):
        if value is not None:
            return Decimal(str(value)).quantize(Decimal(str(round_to)))
        else:
            return None



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2016, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import yaml


# Attributes

class PrintBoolean(yaml.YAMLObject):
    yaml_tag = u"!boolean"


class PrintDatasource(yaml.YAMLObject):
    yaml_tag = u"!datasource"


class PrintFeatures(yaml.YAMLObject):
    yaml_tag = u"!features"


class PrintFloat(yaml.YAMLObject):
    yaml_tag = u"!float"


class PrintInteger(yaml.YAMLObject):
    yaml_tag = u"!integer"


class PrintLegend(yaml.YAMLObject):
    yaml_tag = u"!legend"


class PrintMap(yaml.YAMLObject):
    yaml_tag = u"!map"


class PrintNorthArraw(yaml.YAMLObject):
    yaml_tag = u"!northArrow"


class PrintOverviewMap(yaml.YAMLObject):
    yaml_tag = u"!overviewMap"


class PrintPaging(yaml.YAMLObject):
    yaml_tag = u"!paging"


class PrintScalbar(yaml.YAMLObject):
    yaml_tag = u"!scalebar"


class PrintStaticLayer(yaml.YAMLObject):
    yaml_tag = u"!staticLayer"


class PrintString(yaml.YAMLObject):
    yaml_tag = u"!string"


class PrintStringArray(yaml.YAMLObject):
    yaml_tag = u"!stringArray"


class PrintStyle(yaml.YAMLObject):
    yaml_tag = u"!style"


class PrintTable(yaml.YAMLObject):
    yaml_tag = u"!table"


# Configuration


class PrintAcceptAll(yaml.YAMLObject):
    yaml_tag = u"!acceptAll"


class PrintAlwaysAllowedAssertion(yaml.YAMLObject):
    yaml_tag = u"!alwaysAllowedAssertion"


class PrintAndAssertion(yaml.YAMLObject):
    yaml_tag = u"!andAssertion"


class PrintCertificateStore(yaml.YAMLObject):
    yaml_tag = u"!certificateStore"


class PrintCredential(yaml.YAMLObject):
    yaml_tag = u"!credential"


class PrintDmsMatch(yaml.YAMLObject):
    yaml_tag = u"!dnsMatch"


class PrintIpMatch(yaml.YAMLObject):
    yaml_tag = u"!ipMatch"


class PrintLocalMatch(yaml.YAMLObject):
    yaml_tag = u"!localMatch"


class PrintMergeSource(yaml.YAMLObject):
    yaml_tag = u"!mergeSource"


class PrintOldApi(yaml.YAMLObject):
    yaml_tag = u"!oldApi"


class PrintPdfConfig(yaml.YAMLObject):
    yaml_tag = u"!pdfConfig"


class PrintProxy(yaml.YAMLObject):
    yaml_tag = u"!proxy"


class PrintRoleAccessAssertion(yaml.YAMLObject):
    yaml_tag = u"!roleAccessAssertion"


class PrintTemplate(yaml.YAMLObject):
    yaml_tag = u"!template"


class PrintUpdatePdfConfigUpdate(yaml.YAMLObject):
    yaml_tag = u"!updatePdfConfigUpdate"


class PrintUrlImage(yaml.YAMLObject):
    yaml_tag = u"!urlImage"


class PrintZoomLevels(yaml.YAMLObject):
    yaml_tag = u"!zoomLevels"


# Processor


class PrintAddBackgroundLayers(yaml.YAMLObject):
    yaml_tag = u"!addBackgroundLayers"


class PrintaddHeaders(yaml.YAMLObject):
    yaml_tag = u"!addHeaders"


class PrintaddOverlayLayers(yaml.YAMLObject):
    yaml_tag = u"!addOverlayLayers"


class PrintconfigureHttpRequests(yaml.YAMLObject):
    yaml_tag = u"!configureHttpRequests"


class PrintcreateDataSource(yaml.YAMLObject):
    yaml_tag = u"!createDataSource"


class PrintcreateMap(yaml.YAMLObject):
    yaml_tag = u"!createMap"


class PrintcreateMapPages(yaml.YAMLObject):
    yaml_tag = u"!createMapPages"


class PrintcreateNorthArrow(yaml.YAMLObject):
    yaml_tag = u"!createNorthArrow"


class PrintcreateOverviewMap(yaml.YAMLObject):
    yaml_tag = u"!createOverviewMap"


class PrintcreateScalebar(yaml.YAMLObject):
    yaml_tag = u"!createScalebar"


class PrintForwardHeaders(yaml.YAMLObject):
    yaml_tag = u"!forwardHeaders"


class PrintMapUri(yaml.YAMLObject):
    yaml_tag = u"!mapUri"


class PrintMergeDataSources(yaml.YAMLObject):
    yaml_tag = u"!mergeDataSources"


class PrintPrepareLegend(yaml.YAMLObject):
    yaml_tag = u"!prepareLegend"


class PrintPrepareTable(yaml.YAMLObject):
    yaml_tag = u"!prepareTable"


class PrintReportBuilder(yaml.YAMLObject):
    yaml_tag = u"!reportBuilder"


class PrintRestrictUris(yaml.YAMLObject):
    yaml_tag = u"!restrictUris"


class PrintSetFeatures(yaml.YAMLObject):
    yaml_tag = u"!setFeatures"


class PrintSetStyle(yaml.YAMLObject):
    yaml_tag = u"!setStyle"


class PrintSetWmsCustomParam(yaml.YAMLObject):
    yaml_tag = u"!setWmsCustomParam"


class PrintUpdatePdfConfig(yaml.YAMLObject):
    yaml_tag = u"!updatePdfConfig"


class PrintUseHttpForHttps(yaml.YAMLObject):
    yaml_tag = u"!useHttpForHttps"


# Other


class PrintConfigureHttpRequests(yaml.YAMLObject):
    yaml_tag = u"!configureHttpRequests"


class PrintCreateNorthArrow(yaml.YAMLObject):
    yaml_tag = u"!createNorthArrow"


class PrintCreateScalebar(yaml.YAMLObject):
    yaml_tag = u"!createScalebar"


class PrintCreateMap(yaml.YAMLObject):
    yaml_tag = u"!createMap"


class PrintCreateDataSource(yaml.YAMLObject):
    yaml_tag = u"!createDataSource"



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2016, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import httplib2
import subprocess
import os
import yaml
import re
import traceback
from json import loads
from urlparse import urlsplit
from xml.dom.minidom import parseString
from sqlalchemy.exc import ProgrammingError, NoSuchTableError
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.util import class_mapper
from sqlalchemy.orm.properties import ColumnProperty
from geoalchemy2.types import Geometry

from lingua.extractors import Extractor
from lingua.extractors import Message
from pyramid.paster import bootstrap
from c2c.template import get_config
from bottle import template, MakoTemplate
from mako.template import Template
from mako.lookup import TemplateLookup

from c2cgeoportal.lib import add_url_params, get_url2
from c2cgeoportal.lib.bashcolor import colorize, RED
from c2cgeoportal.lib.dbreflection import get_class
from c2cgeoportal.lib.caching import init_region


class GeoMapfishAngularExtractor(Extractor):  # pragma: no cover
    "GeoMapfish angular extractor"

    extensions = [".js", ".html"]

    def __call__(self, filename, options):
        config = get_config(".build/config.yaml")

        class Registry:
            settings = config

        class Request:
            registry = Registry()
            params = {}
            GET = {}
            user_agent = ""

            def static_url(*args, **kwargs):
                return ""

            def static_path(*args, **kwargs):
                return ""

            def route_url(*args, **kwargs):
                return ""

            def current_route_url(*args, **kwargs):
                return ""

        init_region({"backend": "dogpile.cache.memory"})

        int_filename = filename
        if re.match("^" + re.escape("./{}/templates".format(config["package"])), filename):
            try:
                empty_template = Template("")

                class Lookup(TemplateLookup):
                    @staticmethod
                    def get_template(uri):
                        return empty_template

                class MyTemplate(MakoTemplate):
                    def prepare(self, **options):
                        options.update({"input_encoding": self.encoding})
                        lookup = Lookup(**options)
                        if self.source:
                            self.tpl = Template(self.source, lookup=lookup, **options)
                        else:
                            self.tpl = Template(
                                uri=self.name,
                                filename=self.filename,
                                lookup=lookup, **options)

                processed = template(
                    filename,
                    {
                        "request": Request(),
                        "lang": "fr",
                        "debug": False,
                        "extra_params": {},
                        "permalink_themes": "",
                        "fulltextsearch_groups": [],
                        "wfs_types": [],
                        "_": lambda x: x,
                    },
                    template_adapter=MyTemplate
                )
                int_filename = os.path.join(os.path.dirname(filename), "_" + os.path.basename(filename))
                with open(int_filename, "wb") as file_open:
                    file_open.write(processed.encode("utf-8"))
            except:
                print(traceback.format_exc())

        message_str = subprocess.check_output(["node", "tools/extract-messages.js", int_filename])
        if int_filename != filename:
            os.unlink(int_filename)
        try:
            messages = []
            for contexts, message in loads(message_str):
                for context in contexts.split(", "):
                    messages.append(Message(
                        None, message, None, [], u"", u"", context.split(":")
                    ))
            return messages
        except:
            print(colorize("An error occurred", RED))
            print(colorize(message_str, RED))
            print("------")
            raise


class GeoMapfishConfigExtractor(Extractor):  # pragma: no cover
    "GeoMapfish config extractor (raster layers, and print templates)"

    extensions = [".yaml"]

    def __call__(self, filename, options):
        with open(filename) as config_file:
            from c2cgeoportal.lib.print_ import *  # noqa
            config = yaml.load(config_file)
            # for application config (.build/config.yaml)
            if "vars" in config:
                return [
                    Message(
                        None, raster_layer, None, [], u"", u"",
                        (filename, u"raster/{0!s}".format(raster_layer))
                    )
                    for raster_layer in config["vars"].get("raster", {}).keys()
                ]
            # for the print config
            elif "templates" in config:
                result = []
                for template_ in config.get("templates").keys():
                    result.append(Message(
                        None, template_, None, [], u"", u"",
                        (filename, u"template/{0!s}".format(template_))
                    ))
                    result += [
                        Message(
                            None, attribute, None, [], u"", u"",
                            (filename, u"template/{0!s}/{1!s}".format(template_, attribute))
                        )
                        for attribute in config.get("templates")[template_].attributes.keys()
                    ]
                return result
            else:
                raise Exception("Not a known config file")


class GeoMapfishThemeExtractor(Extractor):  # pragma: no cover
    "GeoMapfish theme extractor"

    # Run on the development.ini file
    extensions = [".ini"]
    featuretype_cache = {}

    def __call__(self, filename, options):
        messages = []

        self.env = bootstrap(filename)
        with open("project.yaml") as f:
            self.package = yaml.load(f)
        self.config = get_config(".build/config.yaml")

        try:
            from c2cgeoportal.models import DBSession, Theme, LayerGroup, \
                LayerWMS, LayerWMTS, FullTextSearch

            self._import(Theme, messages)
            self._import(LayerGroup, messages)
            self._import(LayerWMS, messages, self._import_layer_wms)
            self._import(LayerWMTS, messages, self._import_layer_wmts)
        except ProgrammingError as e:
            print(colorize(
                "ERROR: The database is probably not up to date "
                "(should be ignored when happen during the upgrade)",
                RED
            ))
            print(colorize(e, RED))

        for ln, in DBSession.query(FullTextSearch.layer_name).distinct().all():
            if ln is not None and ln != "":
                messages.append(Message(
                    None, ln, None, [], u"", u"",
                    ("fts", ln.encode("ascii", errors="replace"))
                ))

        return messages

    @staticmethod
    def _import(object_type, messages, callback=None):
        from c2cgeoportal.models import DBSession

        items = DBSession.query(object_type)
        for item in items:
            messages.append(Message(
                None, item.name, None, [], u"", u"",
                (item.item_type, item.name.encode("ascii", errors="replace"))
            ))

            if callback is not None:
                callback(item, messages)

    def _import_layer_wms(self, layer, messages):
        server = layer.ogc_server
        url = server.url_wfs or server.url
        if url is None:
            return
        for wms_layer in layer.layer.split(","):
            self._import_layer_attributes(
                url, wms_layer, layer.item_type, layer.name, layer.id, messages
            )
        if layer.geo_table is not None and layer.geo_table != "":
            exclude = [] if layer.exclude_properties is None else layer.exclude_properties.split(",")
            last_update_date = layer.get_metadatas("lastUpdateDateColumn")
            if len(last_update_date) == 1:
                exclude.append(last_update_date[0].value)
            last_update_user = layer.get_metadatas("lastUpdateUserColumn")
            if len(last_update_user) == 1:
                exclude.append(last_update_user[0].value)
            try:
                cls = get_class(layer.geo_table, exclude_properties=exclude)
                for column_property in class_mapper(cls).iterate_properties:
                    if isinstance(column_property, ColumnProperty) and \
                            len(column_property.columns) == 1 and \
                            not column_property.columns[0].primary_key and \
                            not column_property.columns[0].foreign_keys and \
                            not isinstance(column_property.columns[0].type, Geometry):
                        messages.append(Message(
                            None, column_property.key, None, [], "", "",
                            (".".join(["edit", layer.item_type, str(layer.id)]), layer.name)
                        ))
            except NoSuchTableError:
                exit(colorize("No such table '{}' for layer '{}'.".format(layer.geo_table, layer.name), RED))

    def _import_layer_wmts(self, layer, messages):
        from c2cgeoportal.models import DBSession, OGCServer

        layers = [d.value for d in layer.metadatas if d.name == "queryLayers"]
        if len(layers) == 0:
            layers = [d.value for d in layer.metadatas if d.name == "wmsLayer"]
        server = [d.value for d in layer.metadatas if d.name == "ogcServer"]
        if len(server) >= 1 and len(layers) >= 1:
            for wms_layer in layers:
                try:
                    db_server = DBSession.query(OGCServer).filter(OGCServer.name == server[0]).one()
                    self._import_layer_attributes(
                        db_server.url_wfs or db_server.url, wms_layer,
                        layer.item_type, layer.name, layer.id, messages
                    )
                except NoResultFound:
                    print("Error: the OGC server '{}' from the WMTS layer '{}' does not exist.".format(
                        server[0], layer.name
                    ))

    def _import_layer_attributes(self, url, layer, item_type, name, layer_id, messages):
        for attribute in self._layer_attributes(url, layer):
            messages.append(Message(
                None, attribute, None, [], "", "", (".".join([item_type, name]), layer)
            ))

    def _layer_attributes(self, url, layer):
        errors = set()

        class Registry:
            setting = None

        class Request:
            registry = Registry()

        request = Request()
        request.registry.settings = self.config
        # static schema will not be supported
        url = get_url2("Layer", url, request, errors)
        if len(errors) > 0:
            print("\n".join(errors))
            return []

        url = add_url_params(url, {
            "SERVICE": "WFS",
            "VERSION": "1.1.0",
            "REQUEST": "DescribeFeatureType",
        })

        if url not in self.featuretype_cache:
            print("Get DescribeFeatureType for url: {0!s}".format(url))
            self.featuretype_cache[url] = None

            # forward request to target (without Host Header)
            http = httplib2.Http()
            h = {}
            if urlsplit(url).hostname == "localhost":  # pragma: no cover
                h["Host"] = self.package["host"]
            try:
                resp, content = http.request(url, method="GET", headers=h)
            except:  # pragma: no cover
                print("Unable to DescribeFeatureType from URL {0!s}".format(url))
                self.featuretype_cache[url] = None
                return []

            if resp.status < 200 or resp.status >= 300:  # pragma: no cover
                print("DescribeFeatureType from URL {0!s} return the error: {1:d} {2!s}".format(
                    url, resp.status, resp.reason
                ))
                self.featuretype_cache[url] = None
                return []

            try:
                describe = parseString(content)
                self.featuretype_cache[url] = describe
            except AttributeError:
                print(
                    "WARNING! an error occured while trying to "
                    "read the Mapfile and recover the themes."
                )
                print("URL: {0!s}\nxml:\n{1!s}".format(url, content))
        else:
            describe = self.featuretype_cache[url]

        if describe is None:
            return []

        attributes = []
        # Should probably be adapted for other king of servers
        for type_element in describe.getElementsByTagNameNS(
            "http://www.w3.org/2001/XMLSchema", "complexType"
        ):
            if type_element.getAttribute("name") == "{0!s}Type".format(layer):
                for element in type_element.getElementsByTagNameNS(
                    "http://www.w3.org/2001/XMLSchema", "element"
                ):
                    if not element.getAttribute("type").startswith("gml:"):
                        attributes.append(element.getAttribute("name"))
        return attributes



# -*- coding: utf-8 -*-

# Copyright (c) 2013-2016, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import isodate
import datetime


class TimeInformation:
    """
    Arguments:

    * ``extent`` A time extent instance (``TimeExtentValue`` or
                 ``TimeExtentInterval``)
    * ``mode`` The layer mode ("single", "range" or "disabled")
    * ``widget`` The layer mode ("slider" (default) or "datepicker")
    """
    def __init__(self):
        self.extent = None
        self.mode = None
        self.widget = None
        self.layer = None

    def merge(self, layer, extent, mode, widget):
        layer_apply = \
            self.layer == layer or \
            (not self.has_time() and extent is not None)

        self.merge_extent(extent)
        self.merge_mode(mode)
        self.merge_widget(widget)

        if layer_apply:
            layer["time"] = self.to_dict()
            self.layer = layer
        elif self.layer is not None:
            del self.layer["time"]
            self.layer = None

    def merge_extent(self, extent):
        if self.extent is not None:
            self.extent.merge(extent)
        else:
            self.extent = extent

    def merge_mode(self, mode):
        if mode != "disabled":
            if self.mode is not None:
                if self.mode != mode:
                    raise ValueError(
                        "Could not mix time mode '{0!s}' and '{1!s}'".format(mode, self.mode)
                    )
            else:
                self.mode = mode

    def merge_widget(self, widget):
        widget = "slider" if not widget else widget

        if self.widget is not None:
            if self.widget != widget:
                raise ValueError(
                    "Could not mix time widget '{0!s}' and '{1!s}'".format(widget, self.widget)
                )
        else:
            self.widget = widget

    def has_time(self):
        return self.extent is not None

    def to_dict(self):
        if self.has_time():
            time = self.extent.to_dict()
            time["mode"] = self.mode
            time["widget"] = self.widget
            return time
        else:
            return None


class TimeExtentValue:
    """
    Represents time as a list of values.
    """
    def __init__(self, values, resolution, min_def_value, max_def_value):
        """
        Arguments:

        * ``values`` A set() of datetime
        * ``resolution`` The resolution from the mapfile time definition
        * ``min_def_value`` the minimum default value as a datetime
        * ``max_def_value`` the maximum default value as a datetime
        """
        self.values = values
        self.resolution = resolution
        self.min_def_value = min_def_value
        self.max_def_value = max_def_value

    def merge(self, extent):
        if not isinstance(extent, TimeExtentValue):
            raise ValueError(
                "Could not mix time defined as a list of "
                "values with other type of definition")
        self.values.update(extent.values)
        self.min_def_value = min(self.min_def_value, extent.min_def_value)
        self.max_def_value = max(self.max_def_value, extent.max_def_value)

    def to_dict(self):
        values = sorted(self.values)
        min_def_value = _format_date(self.min_def_value) \
            if self.min_def_value else None
        max_def_value = _format_date(self.max_def_value) \
            if self.max_def_value else None

        return {
            "minValue": _format_date(values[0]),
            "maxValue": _format_date(values[-1]),
            "values": map(_format_date, values),
            "resolution": self.resolution,
            "minDefValue": min_def_value,
            "maxDefValue": max_def_value,
        }


class TimeExtentInterval:
    """
    Represents time with the help of a start, an end and an interval.
    """
    def __init__(self, start, end, interval, resolution,
                 min_def_value, max_def_value):
        """
        Arguments:

        * ``start`` The start value as a datetime
        * ``end`` The end value as a datetime
        * ``interval`` The interval as a tuple (years, months, days, seconds)
        * ``resolution`` The resolution from the mapfile time definition
        * ``min_def_value`` the minimum default value as a datetime
        * ``max_def_value`` the maximum default value as a datetime
        """
        self.start = start
        self.end = end
        self.interval = interval
        self.resolution = resolution
        self.min_def_value = min_def_value
        self.max_def_value = max_def_value

    def merge(self, extent):
        if not isinstance(extent, TimeExtentInterval):
            raise ValueError(
                "Could not merge time defined as with an "
                " interval with other type of definition")
        if self.interval != extent.interval:
            raise ValueError(
                "Could not merge times defined with a "
                "different interval")
        self.start = min(self.start, extent.start)
        self.end = max(self.end, extent.end)
        self.min_def_value = min(self.min_def_value, extent.min_def_value)
        self.max_def_value = max(self.max_def_value, extent.max_def_value)

    def to_dict(self):
        min_def_value = _format_date(self.min_def_value) \
            if self.min_def_value is not None else None
        max_def_value = _format_date(self.max_def_value) \
            if self.max_def_value is not None else None

        return {
            "minValue": _format_date(self.start),
            "maxValue": _format_date(self.end),
            "interval": self.interval,
            "resolution": self.resolution,
            "minDefValue": min_def_value,
            "maxDefValue": max_def_value,
        }


def parse_extent(extent, default_values):
    """
    Parse a time extend from OWSLib to a `̀ TimeExtentValue`` or a
    ``TimeExtentInterval``

    Two formats are supported:
    * ['start/end/interval']
    * ['date1', 'date2', ..., 'date N']

    default_values must be a slash separated String from OWSLib's a
    defaulttimeposition
    """
    if len(extent) > 0:
        min_def_value, max_def_value = _parse_default_values(default_values)
        if extent[0].count("/") > 0:
            # case "start/end/interval"
            if len(extent) > 1 or extent[0].count("/") != 2:
                raise ValueError(
                    "Unsupported time definition '{0!s}'".format(extent))
            s, e, i = extent[0].split("/")
            start = _parse_date(s)
            end = _parse_date(e)
            interval = _parse_duration(i)

            return TimeExtentInterval(start[1], end[1], interval, start[0],
                                      min_def_value, max_def_value)
        else:
            # case "value1, value2, ..., valueN"
            dates = [_parse_date(d) for d in extent]
            resolution = dates[0][0]
            values = set(d[1] for d in dates)

            return TimeExtentValue(values, resolution, min_def_value,
                                   max_def_value)
    else:
        raise ValueError("Invalid time extent format '%s'", extent)


def _parse_default_values(default_values):
    """
    Parse the 'default' value from OWSLib's defaulttimeposition
    and return a maximum of two dates. default value must be a
    slash separated String.
    return None on the seconde value if it doesn't exist.
    """
    if default_values is None:  # pragma: no cover
        return None, None

    def_value = str(default_values).split("/")

    r, min_def_value = _parse_date(def_value[0])
    max_def_value = None

    if len(def_value) > 1:
        r, max_def_value = _parse_date(def_value[1])

    return min_def_value, max_def_value


def _parse_date(date):
    """
    Parses a string into a tuple containing:

    * the resolution: "year", "month", "day" or "second"
    * the date as a datetime

    The returned datetime always has a timezone (default to UTC)
    """
    resolutions = {
        "year": "%Y",
        "month": "%Y-%m",
        "day": "%Y-%m-%d",
    }

    for resolution, pattern in resolutions.items():
        try:
            dt = datetime.datetime.strptime(date, pattern)
            return resolution, dt.replace(tzinfo=isodate.UTC)
        except:
            pass

    try:
        dt = isodate.parse_datetime(date)
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=isodate.UTC)
        return "second", dt
    except:
        raise ValueError("Invalid date format '{0!s}'".format(date))


def _format_date(date):
    assert isinstance(date, datetime.datetime)
    str = isodate.datetime_isoformat(date)
    if date.tzinfo is None:
        str += "Z"
    return str


def _parse_duration(duration):
    """
    Parses an ISO 8601 duration (i.e. "P2DT5S") and returns a tuple containing:

    * years
    * months
    * days
    * seconds
    """
    duration = isodate.parse_duration(duration)

    # casting years and months to int as isodate might return a float
    return (
        int(duration.years) if hasattr(duration, "years") else 0,
        int(duration.months) if hasattr(duration, "months") else 0,
        duration.days,
        duration.seconds,
    )



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2016, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import httplib2
import logging
import json
import sys
import urlparse
import re

from random import Random
from math import sqrt
from xml.dom.minidom import parseString
from socket import gaierror
from collections import Counter

from pyramid.view import view_config
from pyramid.i18n import TranslationStringFactory
from pyramid.httpexceptions import HTTPFound, HTTPBadRequest, HTTPForbidden, HTTPBadGateway
from pyramid.security import remember, forget
from pyramid.response import Response
from sqlalchemy import func
from sqlalchemy.orm.exc import NoResultFound
from owslib.wms import WebMapService

from c2cgeoportal.lib import get_setting, get_protected_layers_query, \
    get_url2, get_url, get_typed, get_types_map, add_url_params
from c2cgeoportal.lib.cacheversion import get_cache_version
from c2cgeoportal.lib.caching import get_region, invalidate_region,  \
    set_common_headers, NO_CACHE, PUBLIC_CACHE, PRIVATE_CACHE
from c2cgeoportal.lib.functionality import get_functionality, \
    get_mapserver_substitution_params
from c2cgeoportal.lib.wmstparsing import parse_extent, TimeInformation
from c2cgeoportal.lib.email_ import send_email
from c2cgeoportal.models import DBSession, User, Role, \
    Theme, LayerGroup, RestrictionArea, Interface, OGCServer, \
    Layer, LayerV1, LayerWMS, LayerWMTS, FullTextSearch, \
    OGCSERVER_TYPE_GEOSERVER, OGCSERVER_TYPE_MAPSERVER, \
    OGCSERVER_AUTH_GEOSERVER, OGCSERVER_AUTH_NOAUTH


_ = TranslationStringFactory("c2cgeoportal")
log = logging.getLogger(__name__)
cache_region = get_region()


class DimensionInformation:

    URL_PART_RE = re.compile("[a-zA-Z0-9_\-~\.]+$")

    def __init__(self):
        self._dimensions = {}

    def merge(self, layer, layer_node, mixed):
        errors = set()

        dimensions = {}
        for dimension in layer.dimensions:
            if dimension.value is not None and not self.URL_PART_RE.match(dimension.value):  # pragma: nocover
                errors.add("The layer '{}' has an unsupported dimension value '{}' ('{}').".format(
                    layer.name, dimension.value, dimension.name
                ))
            elif dimension.name in dimensions:  # pragma: nocover
                errors.add("The layer '{}' has a duplicated dimension name '{}'.".format(
                    layer.name, dimension.name
                ))
            else:
                dimensions[dimension.name] = dimension.value

        if mixed:
            layer_node["dimensions"] = dimensions
        else:
            for name, value in dimensions.items():
                if name not in self._dimensions or self._dimensions[name] is None:
                    self._dimensions[name] = value
                elif self._dimensions[name] != value and value is not None:
                    errors.add(
                        "The layer '{}' has a wrong dimension value '{}' for '{}', "
                        "expected '{}' or empty.".format(
                            layer.name, value, name, self._dimensions[name]
                        )
                    )
        return errors

    def get_dimensions(self):
        return self._dimensions


class Entry:

    WFS_NS = "http://www.opengis.net/wfs"

    def __init__(self, request):
        self.request = request
        self.settings = request.registry.settings
        self.mapserver_settings = self.settings.get("mapserverproxy", {})
        self.debug = "debug" in request.params
        self.lang = request.locale_name
        self.metadata_type = get_types_map(
            self.settings.get("admin_interface", {}).get("available_metadata", [])
        )
        if "default_ogc_server" in self.mapserver_settings:
            try:
                self.default_ogc_server = DBSession.query(OGCServer).filter(
                    OGCServer.name == self.mapserver_settings["default_ogc_server"]
                ).one()
            except NoResultFound:  # pragma: no cover
                log.error("Unable to find the OGC server named: {}.".format(
                    self.mapserver_settings["default_ogc_server"])
                )
                log.error("Available OGC servers: {}".format(
                    ", ".join([i[0] for i in DBSession.query(OGCServer.name).all()]))
                )

        if "external_ogc_server" in self.mapserver_settings:
            try:
                self.external_ogc_server = DBSession.query(OGCServer).filter(
                    OGCServer.name == self.mapserver_settings["external_ogc_server"]
                ).one()
            except NoResultFound:  # pragma: no cover
                log.error("Unable to find the OGC server named: {}.".format(
                    self.mapserver_settings["external_ogc_server"])
                )
                log.error("Available OGC servers: {}".format(
                    ", ".join([i[0] for i in DBSession.query(OGCServer.name).all()]))
                )

    @view_config(route_name="testi18n", renderer="testi18n.html")
    def testi18n(self):  # pragma: no cover
        _ = self.request.translate
        return {"title": _("title i18n")}

    def _get_cache_role_key(self, ogc_server):
        return self._get_role_id() if (
            ogc_server.auth != OGCSERVER_AUTH_NOAUTH
        ) else None

    def _get_capabilities_cache_role_key(self, ogc_server):
        return self._get_role_id() if (
            ogc_server.auth != OGCSERVER_AUTH_NOAUTH and
            ogc_server.type != OGCSERVER_TYPE_MAPSERVER
        ) else None

    def _get_metadatas(self, item, errors):
        metadatas = {}
        for metadata in item.metadatas:
            value = get_typed(
                metadata.name, metadata.value,
                self.metadata_type, self.request, errors
            )
            if value is not None:
                metadatas[metadata.name] = value

        return metadatas

    def _wms_getcap(self, ogc_server=None):
        ogc_server = (ogc_server or self.default_ogc_server)
        errors = set()
        url = get_url2(
            "The OGC server '{}'".format(ogc_server.name),
            ogc_server.url, self.request, errors
        )
        if len(errors):  # pragma: no cover
            return None, errors

        # add functionalities params
        sparams = get_mapserver_substitution_params(self.request)
        url = add_url_params(url, sparams)

        return self._wms_getcap_cached(
            url, ogc_server.auth, self._get_capabilities_cache_role_key(ogc_server)
        )

    @cache_region.cache_on_arguments()
    def _wms_getcap_cached(self, url, auth, role_id):
        """ role_id is just for cache """

        errors = set()
        wms = None

        url = add_url_params(url, {
            "SERVICE": "WMS",
            "VERSION": "1.1.1",
            "REQUEST": "GetCapabilities",
        })

        log.info("Get WMS GetCapabilities for url: {0!s}".format(url))

        # forward request to target (without Host Header)
        http = httplib2.Http()
        headers = dict(self.request.headers)

        role = None if self.request.user is None else self.request.user.role

        # Add headers for Geoserver
        if auth == OGCSERVER_AUTH_GEOSERVER and self.request.user is not None:
            headers["sec-username"] = self.request.user.username
            headers["sec-roles"] = role.name

        if urlparse.urlsplit(url).hostname != "localhost":  # pragma: no cover
            headers.pop("Host")
        try:
            resp, content = http.request(url, method="GET", headers=headers)
        except:  # pragma: no cover
            errors.add("Unable to GetCapabilities from url {0!s}".format(url))
            return None, errors

        if resp.status < 200 or resp.status >= 300:  # pragma: no cover
            error = "GetCapabilities from URL {0!s} return the error: {1:d} {2!s}".format(
                url, resp.status, resp.reason
            )
            errors.add(error)
            log.exception(error)
            return None, errors

        # With wms 1.3 it returns text/xml also in case of error :-(
        if resp.get("content-type").split(";")[0].strip() not in \
                ["application/vnd.ogc.wms_xml", "text/xml"]:
            error = "GetCapabilities from URL {0!s} returns a wrong Content-Type: {1!s}\n{2!s}".format(
                url, resp.get("content-type"), content.encode("utf-8")
            )
            errors.add(error)
            log.exception(error)
            return None, errors

        try:
            wms = WebMapService(None, xml=content)
        except:  # pragma: no cover
            error = _(
                "WARNING! an error occured while trying to "
                "read the mapfile and recover the themes."
            )
            error = "{0!s}\nURL: {1!s}\n{2!s}".format(error, url, content.encode("utf-8"))
            errors.add(error)
            log.exception(error)
        return wms, errors

    @staticmethod
    def _create_layer_query(role_id, version):
        """ Create an SQLAlchemy query for Layer and for the role
            identified to by ``role_id``.
        """

        if version == 1:
            q = DBSession.query(LayerV1)
        else:
            q = DBSession.query(Layer).with_polymorphic(
                [LayerWMS, LayerWMTS]
            )

        q = q.filter(Layer.public.is_(True))
        if role_id is not None:
            q = q.union(get_protected_layers_query(role_id, version=version))

        return q

    def _get_layer_metadata_urls(self, layer):
        metadata_urls = []
        if len(layer.metadataUrls) > 0:
            metadata_urls = layer.metadataUrls
        for childLayer in layer.layers:
            metadata_urls.extend(self._get_layer_metadata_urls(childLayer))
        return metadata_urls

    def _get_layer_resolution_hint_raw(self, layer):
        resolution_hint_min = None
        resolution_hint_max = None
        if layer.scaleHint:
            # scaleHint is based upon a pixel diagonal length whereas we use
            # resolutions based upon a pixel edge length. There is a sqrt(2)
            # ratio between edge and diagonal of a square.
            resolution_hint_min = float(layer.scaleHint["min"]) / sqrt(2)
            resolution_hint_max = float(layer.scaleHint["max"]) / sqrt(2) \
                if layer.scaleHint["max"] != "0" else 999999999
        for childLayer in layer.layers:
            resolution = self._get_layer_resolution_hint_raw(childLayer)
            resolution_hint_min = resolution[0] if resolution_hint_min is None else (
                resolution_hint_min if resolution[0] is None else
                min(resolution_hint_min, resolution[0])
            )
            resolution_hint_max = resolution[1] if resolution_hint_max is None else (
                resolution_hint_max if resolution[1] is None else
                max(resolution_hint_max, resolution[1])
            )

        return (resolution_hint_min, resolution_hint_max)

    def _get_layer_resolution_hint(self, layer):
        resolution_hint_min, resolution_hint_max = self._get_layer_resolution_hint_raw(layer)
        return (
            0.0 if resolution_hint_min is None else resolution_hint_min,
            999999999 if resolution_hint_max is None else resolution_hint_max,
        )

    def _get_child_layers_info_1(self, layer):
        """ Return information about sub layers of a layer.

            Arguments:

            * ``layer`` The layer object in the WMS capabilities.
        """
        child_layers_info = []
        for child_layer in layer.layers:
            resolution = self._get_layer_resolution_hint(child_layer)
            child_layer_info = {
                "name": child_layer.name,
                "minResolutionHint": float("{0:0.2f}".format(resolution[0])),
                "maxResolutionHint": float("{0:0.2f}".format(resolution[1])),
            }
            if hasattr(child_layer, "queryable"):
                child_layer_info["queryable"] = child_layer.queryable
            child_layers_info.append(child_layer_info)
        return child_layers_info

    def _get_child_layers_info(self, layer):
        """ Return information about sub layers of a layer.

            Arguments:

            * ``layer`` The layer object in the WMS capabilities.
        """
        resolution = self._get_layer_resolution_hint(layer)
        layer_info = {
            "name": layer.name,
            "minResolutionHint": float("{0:0.2f}".format(resolution[0])),
            "maxResolutionHint": float("{0:0.2f}".format(resolution[1])),
        }
        layer_info["queryable"] = layer.queryable == 1 \
            if hasattr(layer, "queryable") else True
        return layer_info

    def _layer(self, layer, time=None, dim=None, role_id=None, mixed=True):
        errors = set()
        l = {
            "id": layer.id,
            "name": layer.name if not isinstance(layer, LayerV1) else layer.layer,
            "metadata": self._get_metadatas(layer, errors),
        }
        if re.search("[/?#]", layer.name):  # pragma: no cover
            errors.add("The layer has an unsupported name '{}'.".format(layer.name))
        if isinstance(layer, LayerWMS) and re.search("[/?#]", layer.layer):  # pragma: no cover
            errors.add("The layer has an unsupported layers '{}'.".format(layer.layer))
        if layer.geo_table:
            self._fill_editable(l, layer)
        if mixed:
            assert(time is None)
            time = TimeInformation()
        assert(time is not None)

        if not isinstance(layer, LayerV1):
            errors |= dim.merge(layer, l, mixed)

        if isinstance(layer, LayerV1):
            wms, wms_layers, wms_errors = self._wms_layers(
                self._get_cache_role_key(self.default_ogc_server),
                self.default_ogc_server,
            )
            errors |= wms_errors
            l.update({
                "type": layer.layer_type,
                "public": layer.public,
                "legend": layer.legend,
                "isChecked": layer.is_checked,
                "isLegendExpanded": layer.is_legend_expanded,
            })
            if layer.identifier_attribute_field:
                l["identifierAttribute"] = layer.identifier_attribute_field
            if layer.disclaimer:
                l["disclaimer"] = layer.disclaimer
            if layer.icon:
                l["icon"] = get_url(layer.icon, self.request, errors=errors)
            if layer.kml:
                l["kml"] = get_url(layer.kml, self.request, errors=errors)
            if layer.metadata_url:
                l["metadataURL"] = layer.metadata_url
            if layer.legend_image:
                l["legendImage"] = get_url(layer.legend_image, self.request, errors=errors)

            if layer.layer_type == "internal WMS":
                self._fill_internal_wms(l, layer, wms, wms_layers, errors)
                errors |= self._merge_time(time, l, layer, wms, wms_layers)
            elif layer.layer_type == "external WMS":
                self._fill_external_wms(l, layer, errors)
            elif layer.layer_type == "WMTS":
                self._fill_wmts(l, layer, errors, role_id=role_id)
        elif isinstance(layer, LayerWMS):
            wms, wms_layers, wms_errors = self._wms_layers(
                self._get_cache_role_key(layer.ogc_server),
                layer.ogc_server,
            )
            errors |= wms_errors
            if layer.layer is None or layer.layer == "":
                errors.add("The layer '{}' don't have any layers".format(layer.name))
                return None, errors
            l["type"] = "WMS"
            l["layers"] = layer.layer
            self._fill_wms(l, layer, errors, role_id, mixed=mixed)
            errors |= self._merge_time(time, l, layer, wms, wms_layers)

        elif isinstance(layer, LayerWMTS):
            l["type"] = "WMTS"
            self._fill_wmts(l, layer, errors, version=2)

        return l, errors

    @staticmethod
    def _merge_time(time, l, layer, wms, wms_layers):
        errors = set()
        wmslayer = layer.name if isinstance(layer, LayerV1) else layer.layer
        try:
            if wmslayer in wms_layers:
                wms_layer_obj = wms[wmslayer]

                if wms_layer_obj.timepositions:
                    extent = parse_extent(
                        wms_layer_obj.timepositions,
                        wms_layer_obj.defaulttimeposition
                    )
                    time.merge(l, extent, layer.time_mode, layer.time_widget)

                for child_layer in wms_layer_obj.layers:
                    if child_layer.timepositions:
                        extent = parse_extent(
                            child_layer.timepositions,
                            child_layer.defaulttimeposition
                        )
                        # The time mode comes from the layer group
                        time.merge(l, extent, layer.time_mode, layer.time_widget)

        except ValueError:  # pragma no cover
            errors.add(
                "Error while handling time for layer '{0!s}': {1!s}".format(layer.name, sys.exc_info()[1])
            )

        return errors

    def _fill_editable(self, l, layer):
        if self.request.user:
            c = DBSession.query(RestrictionArea) \
                .filter(RestrictionArea.roles.any(
                    Role.id == self.request.user.role.id)) \
                .filter(RestrictionArea.layers.any(Layer.id == layer.id)) \
                .filter(RestrictionArea.readwrite.is_(True)) \
                .count()
            if c > 0:
                l["editable"] = True

    def _fill_wms(self, l, layer, errors, role_id, mixed):
        wms, wms_layers, wms_errors = self._wms_layers(role_id, layer.ogc_server)
        errors |= wms_errors

        l["imageType"] = layer.ogc_server.image_type
        if layer.style:  # pragma: no cover
            l["style"] = layer.style

        # now look at what's in the WMS capabilities doc
        l["childLayers"] = []
        for layer_name in layer.layer.split(","):
            if layer_name in wms_layers:
                wms_layer_obj = wms[layer_name]
                metadata_urls = self._get_layer_metadata_urls(wms_layer_obj)
                if len(metadata_urls) > 0:  # pragma: no cover
                    if "metadataUrls" not in l:
                        l["metadataUrls"] = metadata_urls
                    else:
                        l["metadataUrls"].extend(metadata_urls)
                if len(wms_layer_obj.layers) == 0:
                    l["childLayers"].append(self._get_child_layers_info(wms_layer_obj))
                else:
                    for child_layer in wms_layer_obj.layers:
                        l["childLayers"].append(self._get_child_layers_info(child_layer))
            else:
                errors.add(
                    "The layer '{}' ({}) is not defined in WMS capabilities from '{}'".format(
                        layer_name, layer.name, layer.ogc_server.name
                    )
                )

        if "minResolutionHint" not in l:
            min_resolutions_hint = [
                l_["minResolutionHint"]
                for l_ in l["childLayers"]
                if "minResolutionHint" in l_
            ]
            if len(min_resolutions_hint) > 0:
                l["minResolutionHint"] = min(min_resolutions_hint)
        if "maxResolutionHint" not in l:
            max_resolutions_hint = [
                l_["maxResolutionHint"]
                for l_ in l["childLayers"]
                if "maxResolutionHint" in l_
            ]
            if len(max_resolutions_hint) > 0:
                l["maxResolutionHint"] = max(max_resolutions_hint)

        if mixed:
            l["ogcServer"] = layer.ogc_server.name

    def _fill_wms_v1(self, l, layer):
        l["imageType"] = layer.image_type
        if layer.legend_rule:
            l["icon"] = add_url_params(self.request.route_url("mapserverproxy"), {
                "SERVICE": "WMS",
                "VERSION": "1.1.1",
                "REQUEST": "GetLegendGraphic",
                "LAYER": layer.name,
                "FORMAT": "image/png",
                "TRANSPARENT": "TRUE",
                "RULE": layer.legend_rule,
            })
        if layer.style:
            l["style"] = layer.style

    @staticmethod
    def _fill_legend_rule_query_string(l, layer, url):
        if layer.legend_rule and url:
            l["icon"] = add_url_params(url, {
                "SERVICE": "WMS",
                "VERSION": "1.1.1",
                "REQUEST": "GetLegendGraphic",
                "LAYER": layer.name,
                "FORMAT": "image/png",
                "TRANSPARENT": "TRUE",
                "RULE": layer.legend_rule,
            })

    def _fill_internal_wms(self, l, layer, wms, wms_layers, errors):
        self._fill_wms_v1(l, layer)

        self._fill_legend_rule_query_string(
            l, layer,
            self.request.route_url("mapserverproxy")
        )

        # this is a leaf, ie. a Mapserver layer
        if layer.min_resolution is not None:
            l["minResolutionHint"] = layer.min_resolution
        if layer.max_resolution is not None:
            l["maxResolutionHint"] = layer.max_resolution

        wmslayer = layer.name
        # now look at what's in the WMS capabilities doc
        if wmslayer in wms_layers:
            wms_layer_obj = wms[wmslayer]
            metadata_urls = self._get_layer_metadata_urls(wms_layer_obj)
            if len(metadata_urls) > 0:
                l["metadataUrls"] = metadata_urls
            resolutions = self._get_layer_resolution_hint(wms_layer_obj)
            if resolutions[0] <= resolutions[1]:
                if "minResolutionHint" not in l:
                    l["minResolutionHint"] = float("{0:0.2f}".format(resolutions[0]))
                if "maxResolutionHint" not in l:
                    l["maxResolutionHint"] = float("{0:0.2f}".format(resolutions[1]))
            l["childLayers"] = self._get_child_layers_info_1(wms_layer_obj)
            if hasattr(wms_layer_obj, "queryable"):
                l["queryable"] = wms_layer_obj.queryable
        else:
            if self.default_ogc_server.type != OGCSERVER_TYPE_GEOSERVER:
                errors.add(
                    "The layer '{0!s}' is not defined in WMS capabilities".format(wmslayer)
                )

    def _fill_external_wms(self, l, layer, errors):
        self._fill_wms_v1(l, layer)
        self._fill_legend_rule_query_string(l, layer, layer.url)

        if layer.min_resolution is not None:
            l["minResolutionHint"] = layer.min_resolution
        if layer.max_resolution is not None:
            l["maxResolutionHint"] = layer.max_resolution

        l["url"] = get_url(layer.url, self.request, errors=errors)
        l["isSingleTile"] = layer.is_single_tile

    def _fill_wmts(self, l, layer, errors, version=1, role_id=None):
        if version == 1:
            l["url"] = get_url(layer.url, self.request, errors=errors)
        else:
            l["url"] = get_url2(
                "The WMTS layer '{}'".format(layer.name),
                layer.url, self.request, errors=errors
            )

        if layer.style:
            l["style"] = layer.style
        if layer.matrix_set:
            l["matrixSet"] = layer.matrix_set

        if version == 1:
            self._fill_wmts_v1(l, layer, errors, role_id)
        else:
            self._fill_wmts_v2(l, layer)

    @staticmethod
    def _fill_wmts_v2(l, layer):
        l["layer"] = layer.layer
        l["imageType"] = layer.image_type

    def _fill_wmts_v1(self, l, layer, errors, role_id):
        if layer.dimensions:
            try:
                l["dimensions"] = json.loads(layer.dimensions)
            except:  # pragma: no cover
                errors.add(u"Unexpected error: '{0!s}' while reading '{1!s}' in layer '{2!s}'".format(
                    sys.exc_info()[0], layer.dimensions, layer.name
                ))

        mapserverproxy_url = self.request.route_url("mapserverproxy")
        wms, wms_layers, wms_errors = self._wms_layers(
            self._get_cache_role_key(self.default_ogc_server), None,
        )
        errors |= wms_errors

        if layer.wms_url:
            l["wmsUrl"] = layer.wms_url
        elif layer.wms_layers or layer.query_layers:
            l["wmsUrl"] = mapserverproxy_url
        if layer.wms_layers:
            l["wmsLayers"] = layer.wms_layers
        elif layer.wms_url:
            l["wmsLayers"] = layer.name
        # needed for external WMTS
        if layer.query_layers == "[]":  # pragma: no cover
            l["queryLayers"] = []

        if layer.min_resolution is not None:
            l["minResolutionHint"] = layer.min_resolution
        if layer.max_resolution is not None:
            l["maxResolutionHint"] = layer.max_resolution

        # if we have associated local WMS layers then look at what's in the
        # WMS capabilities, and add a queryLayers array with the "resolution
        # hint" information.
        if "wmsUrl" in l and l["wmsUrl"] == mapserverproxy_url:

            query_layers = layer.query_layers.strip("[]") \
                if layer.query_layers else l["wmsLayers"]
            l["queryLayers"] = []

            for query_layer in query_layers.split(","):
                if query_layer not in wms_layers:
                    continue
                query_layer_obj = wms[query_layer]

                ql = {"name": query_layer}
                resolutions = self._get_layer_resolution_hint(query_layer_obj)

                if resolutions[0] != float("Inf"):
                    ql["minResolutionHint"] = float(
                        "{0:0.2f}".format(resolutions[0]))
                if resolutions[1] != 0:
                    ql["maxResolutionHint"] = float(
                        "{0:0.2f}".format(resolutions[1]))

                l["queryLayers"].append(ql)

                # FIXME we do not support WMTS layers associated to
                # MapServer layer groups for now.

    @staticmethod
    def _layer_included(tree_item, version):
        if version == 1 and type(tree_item) == LayerV1:
            return True
        if version == 2 and isinstance(tree_item, Layer):
            return type(tree_item) != LayerV1
        return False

    @staticmethod
    def _is_internal_wms(layer):
        return \
            isinstance(layer, LayerV1) and layer.layer_type == "internal WMS"

    def _get_ogc_servers(self, group, depth=1):
        """ Recurse on all children to get unique identifier for each child. """
        ogc_servers = set()

        # escape loop
        if depth > 30:
            log.error("Error: too many recursions with group '{0!s}'".format(group.name))
            return ogc_servers

        # recurse on children
        if isinstance(group, LayerGroup) and group.children > 0:
            for tree_item in group.children:
                ogc_servers.update(self._get_ogc_servers(tree_item, depth=depth + 1))

        if isinstance(group, LayerWMS):
            ogc_servers.add(group.ogc_server)

        if isinstance(group, LayerWMTS):
            ogc_servers.add(False)

        return ogc_servers

    def _group(
        self, path, group, layers, depth=1, min_levels=1,
        catalogue=True, role_id=None, version=1, mixed=True, time=None, dim=None,
        wms_layers=None, layers_name=None, **kwargs
    ):
        if wms_layers is None:
            wms_layers = []
        if layers_name is None:
            layers_name = []
        children = []
        errors = set()

        if re.search("[/?#]", group.name):  # pragma: no cover
            errors.add("The group has an unsupported name '{}'.".format(group.name))

        # escape loop
        if depth > 30:
            errors.add(
                "Too many recursions with group '{0!s}'".format(group.name)
            )
            return None, errors

        ogc_servers = None
        org_depth = depth
        if depth == 1:
            ogc_servers = list(self._get_ogc_servers(group))
            # check if mixed content
            mixed = len(ogc_servers) != 1 or ogc_servers[0] is False
            if not mixed:
                time = TimeInformation()
            dim = DimensionInformation()

        for tree_item in group.children:
            if type(tree_item) == LayerGroup:
                depth += 1
                if type(group) == Theme or catalogue or \
                        group.is_internal_wms == tree_item.is_internal_wms:
                    gp, gp_errors = self._group(
                        "{0!s}/{1!s}".format(path, tree_item.name),
                        tree_item, layers, depth=depth, min_levels=min_levels,
                        catalogue=catalogue, role_id=role_id, version=version, mixed=mixed,
                        time=time, dim=dim, wms_layers=wms_layers, layers_name=layers_name, **kwargs
                    )
                    errors |= gp_errors
                    if gp is not None:
                        children.append(gp)
                else:
                    errors.add("Group '{0!s}' cannot be in group '{1!s}' (internal/external mix).".format(
                        tree_item.name, group.name
                    ))
            elif self._layer_included(tree_item, version):
                if (tree_item.name in layers):
                    if (catalogue or not isinstance(tree_item, LayerV1) or
                        (isinstance(tree_item, LayerV1) and group.is_internal_wms ==
                            self._is_internal_wms(tree_item))):

                        layers_name.append(tree_item.name)
                        if isinstance(tree_item, LayerWMS):
                            wms_layers.extend(tree_item.layer.split(","))

                        l, l_errors = self._layer(
                            tree_item, role_id=role_id, mixed=mixed, time=time, dim=dim, **kwargs
                        )
                        errors |= l_errors
                        if l is not None:
                            if depth < min_levels:
                                errors.add("The Layer '{0!s}' is under indented ({1:d}/{2:d}).".format(
                                    path + "/" + tree_item.name, depth, min_levels
                                ))
                            else:
                                children.append(l)
                    else:
                        errors.add(
                            "Layer '{0!s}' cannot be in the group '{1!s}' (internal/external mix).".format(
                                tree_item.name, group.name
                            )
                        )

        if len(children) > 0:
            g = {
                "id": group.id,
                "name": group.name,
                "children": children,
                "metadata": self._get_metadatas(group, errors),
                "mixed": False,
            }
            if version == 1:
                g.update({
                    "isExpanded": group.is_expanded,
                    "isInternalWMS": group.is_internal_wms,
                    "isBaseLayer": group.is_base_layer,
                })
                if time is not None and time.has_time() and time.layer is None:
                    g["time"] = time.to_dict()
            else:
                if not mixed:
                    for name, nb in Counter(layers_name).items():
                        if nb > 1:
                            errors.add(
                                "The GeoMapFish layer name '{}', can't be two times "
                                "in the same block (first level group).".format(name)
                            )

                g["mixed"] = mixed
                if org_depth == 1:
                    if not mixed:
                        g["ogcServer"] = ogc_servers[0].name
                        if time.has_time() and time.layer is None:
                            g["time"] = time.to_dict()

                        g["dimensions"] = dim.get_dimensions()

            if version == 1 and group.metadata_url:
                g["metadataURL"] = group.metadata_url

            return g, errors
        else:
            return None, errors

    @cache_region.cache_on_arguments()
    def _layers(self, role_id, version, interface):
        query = self._create_layer_query(role_id, version)
        if interface is not None:
            query = query.join(Layer.interfaces)
            query = query.filter(Interface.name == interface)
        return [l.name for l in query.all()]

    @cache_region.cache_on_arguments()
    def _wms_layers(self, role_id, ogc_server):
        """ role_id is just for cache """

        # retrieve layers metadata via GetCapabilities
        wms, wms_errors = self._wms_getcap(ogc_server)
        if len(wms_errors) > 0:
            return None, [], wms_errors

        return wms, list(wms.contents), wms_errors

    @cache_region.cache_on_arguments()
    def _themes(
        self, role_id, interface="desktop", filter_themes=True, version=1,
        catalogue=False, min_levels=1
    ):
        """
        This function returns theme information for the role identified
        by ``role_id``.
        """
        errors = set()
        layers = self._layers(role_id, version, interface)

        themes = DBSession.query(Theme)
        themes = themes.filter(Theme.public.is_(True))
        if role_id is not None:
            auth_themes = DBSession.query(Theme)
            auth_themes = auth_themes.filter(Theme.public.is_(False))
            auth_themes = auth_themes.join(Theme.restricted_roles)
            auth_themes = auth_themes.filter(Role.id == role_id)

            themes = themes.union(auth_themes)

        themes = themes.order_by(Theme.ordering.asc())

        if filter_themes and interface is not None:
            themes = themes.join(Theme.interfaces)
            themes = themes.filter(Interface.name == interface)

        export_themes = []
        for theme in themes.all():
            if re.search("[/?#]", theme.name):
                errors.add("The theme has an unsupported name '{}'.".format(theme.name))
                continue

            children, children_errors = self._get_children(
                theme, layers, version, catalogue, min_levels, role_id
            )
            errors |= children_errors

            # test if the theme is visible for the current user
            if len(children) > 0:
                icon = get_url2(
                    u"The Theme '{}'".format(theme.name),
                    theme.icon, self.request, errors,
                ) if theme.icon is not None and len(theme.icon) > 0 else self.request.static_url(
                    "c2cgeoportal:static/images/blank.gif"
                )

                t = {
                    "id": theme.id,
                    "name": theme.name,
                    "icon": icon,
                    "children": children,
                    "functionalities": self._get_functionalities(theme),
                    "metadata": self._get_metadatas(theme, errors),
                }
                if version == 1:
                    t.update({
                        "in_mobile_viewer": theme.is_in_interface("mobile"),
                    })
                export_themes.append(t)

        return export_themes, errors

    @staticmethod
    def _get_functionalities(theme):
        result = {}
        for functionality in theme.functionalities:
            if functionality.name in result:
                result[functionality.name].append(functionality.value)
            else:
                result[functionality.name] = [functionality.value]
        return result

    @view_config(route_name="invalidate", renderer="json")
    def invalidate_cache(self):  # pragma: no cover
        invalidate_region()
        return {
            "success": True
        }

    @cache_region.cache_on_arguments()
    def _get_children(self, theme, layers, version, catalogue,
                      min_levels, role_id):
        children = []
        errors = set()
        for item in theme.children:
            if type(item) == LayerGroup:
                gp, gp_errors = self._group(
                    "{0!s}/{1!s}".format(theme.name, item.name),
                    item, layers,
                    role_id=role_id, version=version, catalogue=catalogue,
                    min_levels=min_levels
                )
                errors |= gp_errors
                if gp is not None:
                    children.append(gp)
            elif self._layer_included(item, version):
                if min_levels > 0:
                    errors.add("The Layer '{0!s}' cannot be directly in the theme '{1!s}' (0/{2:d}).".format(
                        item.name, theme.name, min_levels
                    ))
                elif item.name in layers:
                    l, l_errors = self._layer(
                        item, role_id=role_id, dim=DimensionInformation()
                    )
                    errors |= l_errors
                    if l is not None:
                        children.append(l)
        return children, errors

    def _get_wfs_url(self):
        errors = set()
        ogc_server = self.default_ogc_server
        url = get_url2(
            "The OGC server '{}'".format(ogc_server.name),
            ogc_server.url_wfs or ogc_server.url,
            self.request, errors=errors,
        )
        return url, errors

    def _internal_wfs_types(self, role_id):
        url, errors = self._get_wfs_url()
        if len(errors) > 0:  # pragma: no cover
            return None, errors
        return self._wfs_types(url, role_id)

    def _get_external_wfs_url(self):
        errors = set()
        ogc_server = self.external_ogc_server
        url = get_url2(
            "The OGC server '{}'".format(ogc_server.name),
            ogc_server.url_wfs or ogc_server.url,
            self.request, errors=errors,
        )
        return url, errors

    def _external_wfs_types(self, role_id):
        url, errors = self._get_external_wfs_url()
        if len(errors) > 0:  # pragma: no cover
            return None, errors
        return self._wfs_types(url, role_id)

    def _wfs_types(self, wfs_url, role_id):
        # add functionalities query_string
        sparams = get_mapserver_substitution_params(self.request)
        wfs_url = add_url_params(wfs_url, sparams)

        if role_id is not None:
            wfs_url = add_url_params(wfs_url, {"role_id": str(role_id)})

        return self._wfs_types_cached(wfs_url)

    @cache_region.cache_on_arguments()
    def _wfs_types_cached(self, wfs_url):
        errors = set()

        # retrieve layers metadata via GetCapabilities
        params = {
            "SERVICE": "WFS",
            "VERSION": "1.0.0",
            "REQUEST": "GetCapabilities",
        }
        wfsgc_url = add_url_params(wfs_url, params)

        log.info("WFS GetCapabilities for base url: {0!s}".format(wfsgc_url))

        # forward request to target (without Host Header)
        http = httplib2.Http()
        h = dict(self.request.headers)
        if urlparse.urlsplit(wfsgc_url).hostname != "localhost":  # pragma: no cover
            h.pop("Host")
        try:
            resp, get_capabilities_xml = http.request(wfsgc_url, method="GET", headers=h)
        except:  # pragma: no cover
            errors.add("Unable to GetCapabilities from url {0!s}".format(wfsgc_url))
            return None, errors

        if resp.status < 200 or resp.status >= 300:  # pragma: no cover
            errors.add("GetCapabilities from url {0!s} return the error: {1:d} {2!s}".format(
                wfsgc_url, resp.status, resp.reason
            ))
            return None, errors

        try:
            get_capabilities_dom = parseString(get_capabilities_xml)
            featuretypes = []
            for featureType in get_capabilities_dom.getElementsByTagNameNS(
                    self.WFS_NS, "FeatureType"):
                # don't includes FeatureType without name
                name = featureType.getElementsByTagNameNS(self.WFS_NS, "Name").item(0)
                if name:
                    name_value = name.childNodes[0].data
                    # ignore namespace when not using geoserver
                    if name_value.find(":") >= 0 and \
                            not self.default_ogc_server.type == OGCSERVER_TYPE_GEOSERVER:  # pragma nocover
                        name_value = name_value.split(":")[1]
                    featuretypes.append(name_value)
                else:  # pragma nocover
                    log.warn("Feature type without name: {0!s}".format(featureType.toxml()))
            return featuretypes, errors
        except:  # pragma: no cover
            return get_capabilities_xml, errors

    def _external_themes(self, interface):  # pragma nocover
        if not (
            "external_themes_url" in self.settings and
            self.settings["external_themes_url"]
        ):
            return None, set()

        role_id = None
        if self.request.user is not None and \
                hasattr(self.request.user, "parent_role") and \
                self.request.user.parent_role is not None:
            role_id = str(self.request.user.parent_role.id)

        return self._external_themes_role(interface, role_id)

    @cache_region.cache_on_arguments()
    def _external_themes_role(self, interface, role_id):  # pragma nocover
        errors = set()

        ext_url = self.settings["external_themes_url"]
        url_params = {
            "interface": interface
        }

        if ext_url[-1] not in ("?", "&"):
            ext_url += "?"
        ext_url += "&".join([
            "=".join(p) for p in url_params.items()
        ])

        # forward request to target (without Host Header)
        http = httplib2.Http()
        h = dict(self.request.headers)
        if urlparse.urlsplit(ext_url).hostname != "localhost":
            h.pop("Host")
        try:
            resp, content = http.request(ext_url, method="GET", headers=h)
        except:
            errors.add(
                "Unable to get external themes from url {0!s}".format(ext_url)
            )
            return None, errors

        if resp.status < 200 or resp.status >= 300:
            errors.add("Get external themes from url {0!s} return the error: {1:d} {2!s}".format(
                ext_url, resp.status, resp.reason
            ))
            return None, errors

        return content, errors

    def _functionality(self):
        return self._functionality_cached(
            self.request.user.role.name if self.request.user is not None else None
        )

    @cache_region.cache_on_arguments()
    def _functionality_cached(self, role):
        functionality = {}
        for func_ in get_setting(
                self.settings,
                ("functionalities", "available_in_templates"), []
        ):
            functionality[func_] = get_functionality(
                func_, self.request
            )
        return functionality

    @cache_region.cache_on_arguments()
    def _get_layers_enum(self):
        layers_enum = {}
        if "enum" in self.settings.get("layers", {}):
            for layer_name, layer in \
                    self.settings["layers"]["enum"].items():
                layer_enum = {}
                layers_enum[layer_name] = layer_enum
                for attribute in layer["attributes"].keys():
                    layer_enum[attribute] = self.request.route_url(
                        "layers_enumerate_attribute_values",
                        layer_name=layer_name,
                        field_name=attribute,
                        path=""
                    )
        return layers_enum

    def get_cgxp_index_vars(self, templates_params=None):
        if templates_params is None:
            templates_params = {}
        extra_params = {}

        if self.lang:
            extra_params["lang"] = self.lang

        # specific permalink_themes handling
        if "permalink_themes" in templates_params:
            extra_params["permalink_themes"] = templates_params["permalink_themes"]

        d = {
            "lang": self.lang,
            "debug": self.debug,
            "extra_params": extra_params,
        }

        # general templates_params handling
        d.update(templates_params)

        set_common_headers(self.request, "cgxp_index", NO_CACHE)
        return d

    def get_cgxp_permalinktheme_vars(self):
        # call home with extra params
        return self.get_cgxp_index_vars({
            # recover themes from url route
            "permalink_themes": self.request.matchdict["themes"]
        })

    def _get_role_id(self):
        return None if self.request.user is None or self.request.user.role is None else \
            self.request.user.role.id

    def get_cgxp_viewer_vars(self):
        role_id = self._get_role_id()
        interface = self.request.interface_name

        themes, errors = self._themes(role_id, interface)
        wfs_types, add_errors = self._internal_wfs_types(role_id)
        errors |= add_errors

        version_params = {
            "cache_version": get_cache_version()
        }
        version_role_params = {
            "cache_version": get_cache_version()
        }
        if role_id is not None:
            version_role_params["user"] = role_id

        d = {
            "lang": self.lang,
            "debug": self.debug,
            "themes": json.dumps(themes),
            "user": self.request.user,
            "WFSTypes": json.dumps(wfs_types),
            "tiles_url": json.dumps(self.settings.get("tiles_url")),
            "functionality": self._functionality(),
            "queryer_attribute_urls": json.dumps(self._get_layers_enum()),
            "serverError": json.dumps(list(errors)),
            "version_params": version_params,
            "version_role_params": version_role_params,
        }

        if hasattr(self, "external_ogc_server"):
            external_wfs_types, add_errors = self._external_wfs_types(role_id)
            errors |= add_errors
            d["externalWFSTypes"] = json.dumps(external_wfs_types)

            external_themes, add_errors = self._external_themes(interface)
            errors |= add_errors
            d["external_themes"] = external_themes

        # handle permalink_themes
        permalink_themes = self.request.params.get("permalink_themes")
        if permalink_themes:
            d["permalink_themes"] = json.dumps(permalink_themes.split(","))

        set_common_headers(
            self.request, "cgxp_viewer", NO_CACHE,
            vary=True, content_type="application/javascript",
        )

        return d

    def get_ngeo_index_vars(self, vars=None):
        if vars is None:
            vars = {}
        set_common_headers(self.request, "ngeo_index", NO_CACHE)

        vars["debug"] = self.debug

        vars["fulltextsearch_groups"] = [
            group[0] for group in DBSession.query(
                func.distinct(FullTextSearch.layer_name)
            ).filter(FullTextSearch.layer_name.isnot(None)).all()
        ]

        url, add_errors = self._get_wfs_url()
        if len(add_errors) > 0:  # pragma: no cover
            wfs_types = None
        else:
            wfs_types, add_errors = self._wfs_types_cached(url)
        if len(add_errors) == 0:
            vars["wfs_types"] = [{
                "featureType": t,
                "label": t,
            } for t in wfs_types]
        else:  # pragma: no cover
            log.error("Error while getting the WFS params: \n{}".format("\n".join(add_errors)))
            vars["wfs_types"] = []

        return vars

    def get_ngeo_permalinktheme_vars(self):
        # recover themes from url route
        themes = self.request.matchdict["themes"]
        d = {}
        d["permalink_themes"] = themes
        # call home with extra params
        return self.get_ngeo_index_vars(d)

    @view_config(route_name="apijs", renderer="api/api.js")
    def apijs(self):
        wms, wms_errors = self._wms_getcap()
        if len(wms_errors) > 0:  # pragma: no cover
            raise HTTPBadGateway("\n".join(wms_errors))
        queryable_layers = [
            name for name in list(wms.contents)
            if wms[name].queryable == 1]
        cache_version = self.settings.get("cache_version")

        set_common_headers(
            self.request, "apijs", NO_CACHE,
            content_type="application/javascript",
        )

        return {
            "lang": self.lang,
            "debug": self.debug,
            "queryable_layers": json.dumps(queryable_layers),
            "url_params": {"cache_version": cache_version} if cache_version else {},
            "tiles_url": json.dumps(self.settings.get("tiles_url")),
        }

    @view_config(route_name="xapijs", renderer="api/xapi.js")
    def xapijs(self):
        wms, wms_errors = self._wms_getcap()
        if len(wms_errors) > 0:  # pragma: no cover
            raise HTTPBadGateway("\n".join(wms_errors))
        queryable_layers = [
            name for name in list(wms.contents)
            if wms[name].queryable == 1
        ]
        cache_version = self.settings.get("cache_version")

        set_common_headers(
            self.request, "apijs", NO_CACHE,
            content_type="application/javascript",
        )

        return {
            "lang": self.lang,
            "debug": self.debug,
            "queryable_layers": json.dumps(queryable_layers),
            "url_params": {"cache_version": cache_version} if cache_version else {},
            "tiles_url": json.dumps(self.settings.get("tiles_url")),
        }

    @view_config(route_name="apihelp", renderer="api/apihelp.html")
    def apihelp(self):
        set_common_headers(self.request, "apihelp", NO_CACHE)

        return {
            "lang": self.lang,
            "debug": self.debug,
        }

    @view_config(route_name="xapihelp", renderer="api/xapihelp.html")
    def xapihelp(self):
        set_common_headers(self.request, "apihelp", NO_CACHE)

        return {
            "lang": self.lang,
            "debug": self.debug,
        }

    @view_config(route_name="themes", renderer="json")
    def themes(self):
        role_id = self.request.params.get("role_id")
        if role_id is None and self.request.user is not None:
            role_id = self.request.user.role.id
        elif self.request.client_addr != "127.0.0.1":
            role_id = None

        interface = self.request.params.get("interface", "desktop")
        sets = self.request.params.get("set", "all")
        version = int(self.request.params.get("version", 1))
        catalogue = self.request.params.get("catalogue", "false") == "true"
        min_levels = int(self.request.params.get("min_levels", 1))
        group = self.request.params.get("group")
        background_layers_group = self.request.params.get("background")

        export_themes = sets in ("all", "themes")
        export_group = group is not None and sets in ("all", "group")
        export_background = background_layers_group is not None and sets in ("all", "background")

        set_common_headers(self.request, "themes", PRIVATE_CACHE)

        result = {}
        all_errors = set()
        if version == 2:

            result["ogcServers"] = {}
            for ogc_server in DBSession.query(OGCServer).all():
                # required to do everytime to validate the url.
                if ogc_server.auth != OGCSERVER_AUTH_NOAUTH:
                    url = self.request.route_url("mapserverproxy", _query={"ogcserver": ogc_server.name})
                    url_wfs = url
                else:
                    url = get_url2(
                        "The OGC server '{}'".format(ogc_server.name),
                        ogc_server.url, self.request, errors=all_errors
                    )
                    url_wfs = get_url2(
                        "The OGC server (WFS) '{}'".format(ogc_server.name),
                        ogc_server.url_wfs, self.request, errors=all_errors
                    ) if ogc_server.url_wfs is not None else url
                result["ogcServers"][ogc_server.name] = {
                    "url": url,
                    "urlWfs": url_wfs,
                    "type": ogc_server.type,
                    "imageType": ogc_server.image_type,
                    "wfsSupport": ogc_server.wfs_support,
                    "isSingleTile": ogc_server.is_single_tile,
                }
        if export_themes:
            themes, errors = self._themes(
                role_id, interface, True, version, catalogue, min_levels
            )

            if version == 1:
                return themes

            result["themes"] = themes
            all_errors |= errors

        if export_group:
            group, errors = self._get_group(group, role_id, interface, version)
            if group is not None:
                result["group"] = group
            all_errors |= errors

        if export_background:
            group, errors = self._get_group(background_layers_group, role_id, interface, version)
            result["background_layers"] = group["children"] if group is not None else []
            all_errors |= errors

        result["errors"] = list(all_errors)
        if len(all_errors) > 0:
            log.info("Theme errors:\n{}".format("\n".join(all_errors)))
        return result

    def _get_group(self, group, role_id, interface, version):
        layers = self._layers(role_id, version, interface)
        try:
            lg = DBSession.query(LayerGroup).filter(LayerGroup.name == group).one()
            return self._group(
                lg.name, lg, layers,
                role_id=role_id, version=version
            )
        except NoResultFound:  # pragma: no cover
            return None, set(["Unable to find the Group named: {}, Available Groups: {}".format(
                group, ", ".join([i[0] for i in DBSession.query(LayerGroup.name).all()])
            )])

    @view_config(context=HTTPForbidden, renderer="login.html")
    def loginform403(self):
        if self.request.authenticated_userid:
            return HTTPForbidden()  # pragma: no cover

        set_common_headers(self.request, "login", NO_CACHE)

        return {
            "lang": self.lang,
            "came_from": self.request.path,
        }

    @view_config(route_name="loginform", renderer="login.html")
    def loginform(self):
        set_common_headers(self.request, "login", PUBLIC_CACHE, vary=True)

        return {
            "lang": self.lang,
            "came_from": self.request.params.get("came_from") or "/",
        }

    @view_config(route_name="login")
    def login(self):
        login = self.request.POST.get("login")
        password = self.request.POST.get("password")
        if login is None or password is None:  # pragma nocover
            log.info(
                "'login' and 'password' should be "
                "available in request params."
            )
            raise HTTPBadRequest("See server logs for details")
        user = self.request.registry.validate_user(self.request, login, password)
        if user is not None:
            headers = remember(self.request, user)
            log.info("User '{0!s}' logged in.".format(login))

            came_from = self.request.params.get("came_from")
            if came_from:
                return HTTPFound(location=came_from, headers=headers)
            else:
                headers.append(("Content-Type", "text/json"))
                return set_common_headers(
                    self.request, "login", NO_CACHE,
                    response=Response(json.dumps(self._user(
                        DBSession.query(User).filter(User.username == user).one()
                    )), headers=headers),
                )
        else:
            log.info("bad credentials for login '{0!s}'.".format(login))
            raise HTTPBadRequest("See server logs for details")

    @view_config(route_name="logout")
    def logout(self):
        headers = forget(self.request)

        # if there's no user to log out, we send a 404 Not Found (which
        # is the status code that applies best here)
        if not self.request.user:
            log.info("Logout on non login user.")
            raise HTTPBadRequest("See server logs for details")

        log.info("User '{0!s}' ({1:d}) logging out.".format(
            self.request.user.username,
            self.request.user.id
        ))

        headers.append(("Content-Type", "text/json"))
        return set_common_headers(
            self.request, "login", NO_CACHE,
            response=Response("true", headers=headers),
        )

    def _user(self, user=None):
        user = self.request.user if user is None else user
        result = {
            "success": True,  # for Extjs
            "username": user.username,
            "is_password_changed": user.is_password_changed,
            "role_name": user.role_name,
            "role_id": user.role.id
        } if user else {}

        result["functionalities"] = self._functionality()

        return result

    @view_config(route_name="loginuser", renderer="json")
    def loginuser(self):
        set_common_headers(self.request, "login", NO_CACHE)

        return self._user()

    @view_config(route_name="loginchange", renderer="json")
    def loginchange(self):
        set_common_headers(self.request, "login", NO_CACHE)

        old_password = self.request.POST.get("oldPassword")
        new_password = self.request.POST.get("newPassword")
        new_password_confirm = self.request.POST.get("confirmNewPassword")
        if new_password is None or new_password_confirm is None or old_password is None:
            log.info(
                "'oldPassword', 'newPassword' and 'confirmNewPassword' should be "
                "available in request params."
            )
            raise HTTPBadRequest("See server logs for details")

        # check if logged in
        if not self.request.user:  # pragma nocover
            log.info("Change password on non login user.")
            raise HTTPBadRequest("See server logs for details")

        user = self.request.registry.validate_user(
            self.request, self.request.user.username, old_password
        )
        if user is None:
            log.info("The old password is wrong for user '{0!s}'.".format(user))
            raise HTTPBadRequest("See server logs for details")

        if new_password != new_password_confirm:
            log.info(
                "The new password and the new password "
                "confirmation don't match for user '%s'." % user
            )
            raise HTTPBadRequest("See server logs for details")

        u = self.request.user
        u._set_password(new_password)
        u.is_password_changed = True
        DBSession.flush()
        log.info("Password changed for user '{0!s}'".format(self.request.user.username))

        return {
            "success": "true"
        }

    @staticmethod
    def generate_password():
        allchars = "123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
        rand = Random()

        password = ""
        for i in range(8):
            password += rand.choice(allchars)

        return password

    def _loginresetpassword(self):
        username = self.request.POST["login"]
        try:
            user = DBSession.query(User).filter(User.username == username).one()
        except NoResultFound:  # pragma: no cover
            return None, None, None, "The login '{0!s}' doesn't exist.".format(username)

        if user.email is None or user.email == "":  # pragma: no cover
            return \
                None, None, None, \
                "The user '{0!s}' has no registered email address.".format(user.username),

        password = self.generate_password()
        user.set_temp_password(password)

        return user, username, password, None

    @view_config(route_name="loginresetpassword", renderer="json")
    def loginresetpassword(self):  # pragma: no cover
        set_common_headers(
            self.request, "login", NO_CACHE
        )

        user, username, password, error = self._loginresetpassword()
        if error is not None:
            log.info(error)
            raise HTTPBadRequest("See server logs for details")
        settings = self.request.registry.settings["reset_password"]
        try:
            send_email(
                settings["email_from"], [user.email],
                settings["email_body"].format(user=username, password=password).encode("utf-8"),
                settings["email_subject"],
                settings["smtp_server"],
            )

            return {
                "success": True
            }
        except gaierror as e:
            log.error("Unable to send the email.")
            log.exception(e)
            raise HTTPBadRequest("See server logs for details")



# -*- coding: utf-8 -*-

# Copyright (c) 2012-2016, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import shputils
from os.path import dirname
from struct import unpack


class Tile:
    def __init__(self, min_x, min_y, max_x, max_y, filename):
        self.min_x = min_x
        self.min_y = min_y
        self.max_x = max_x
        self.max_y = max_y
        self.filename = filename

    def contains(self, x, y):
        return self.min_x <= x < self.max_x and self.min_y <= y < self.max_y

    def __str__(self):
        return "{0:f}, {1:f}, {2:f}, {3:f}: {4!s}".format(
            self.min_x, self.min_y, self.max_x, self.max_y, self.filename
        )


class BTTile(Tile):
    def get_value(self, x, y):
        file = open(self.filename, "rb")
        if not hasattr(self, "cols"):
            file.seek(10)
            (self.cols, self.rows, self.dataSize, self.floatingPoint) = \
                unpack("<LLhh", file.read(12))
            self.resolution_x = (self.max_x - self.min_x) / self.cols
            self.resolution_y = (self.max_y - self.min_y) / self.rows

        pos_x = int((x - self.min_x) / self.resolution_x)
        pos_y = int((y - self.min_y) / self.resolution_y)
        file.seek(256 + (pos_y + pos_x * self.rows) * self.dataSize)

        if self.floatingPoint == 1:
            val = unpack("<f", file.read(self.dataSize))[0]
        else:
            if self.dataSize == 2:
                format = "<h"
            else:
                format = "<l"
            data = file.read(self.dataSize)
            val = unpack(format, data)[0]

        file.close()
        return val


class GeoRaster:
    def __init__(self, shapefile_name):
        self.tiles = []
        shp_records = shputils.load_shapefile(shapefile_name)
        dir = dirname(shapefile_name)
        if dir == "":
            dir = "."
        for shape in shp_records:
            filename = shape["dbf_data"]["location"].rstrip()
            tile_class = None
            if filename.endswith(".bt"):
                tile_class = BTTile
            if not filename.startswith("/"):
                filename = dir + "/" + filename
            geo = shape["shp_data"]
            tile = tile_class(geo["xmin"], geo["ymin"], geo["xmax"], geo["ymax"], filename)
            self.tiles.append(tile)

    def get_value(self, x, y):
        tile = self._get_tile(x, y)
        if tile:
            return tile.get_value(x, y)
        else:
            return None

    def _get_tile(self, x, y):
        for cur in self.tiles:
            if cur.contains(x, y):
                return cur
        return None



# Copyright (c) 2013-2016, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
with open("README.rst") as f:
    text = f.read().split("\n")
    text = text[3:]
    text += ""
    text += "`Sources <https://github.com/camptocamp/c2cgeoportal/>`_"
    README = "\n".join(text)

install_requires = [
    "pyramid_multiauth",
    "pyramid_mako",  # to render the HTML files
    "psycopg2",
    "GeoAlchemy2",
    "SQLAHelper",
    "pyramid_tm",
    "papyrus",
    "ipcalc",
    "httplib2",
    # Needed by the developement.ini
    "pyramid_debugtoolbar",
    "OWSLib>=0.6.0",
    "dogpile.cache>=0.6",
    "Paste",
    "PasteDeploy",
    "PasteScript",
    # Needed by the production.ini
    "waitress",
    # WMST support
    "isodate",
    "pyramid_closure",
    "lingua",
    "PyYAML",
    "c2c.template>=1.3.0.dev",
]
install_requires += [
    requirement for requirement in
    open(os.path.join(here, "fixversions.txt")).read().splitlines()
    if len(requirement) > 0 and requirement[0] != "#"
]

setup_requires = [
]

tests_require = install_requires + [
]

setup(
    name="c2cgeoportal",
    version="2.1.0.dev23",
    description="c2cgeoportal",
    long_description=README,
    classifiers=[
        "Programming Language :: Python",
        "Framework :: Pylons",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
    ],
    author="Camptocamp",
    author_email="info@camptocamp.com",
    url="http://www.camptocamp.com/geospatial-solutions",
    keywords="web gis geoportail c2cgeoportal geocommune pyramid",
    packages=find_packages(exclude=["*.tests", "*.tests.*"]),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    test_suite="c2cgeoportal",
    entry_points={
        "console_scripts": [
            "print_tpl = c2cgeoportal.scripts.print_tpl:main",
            "manage_users = c2cgeoportal.scripts.manage_users:main",
            "c2ctool = c2cgeoportal.scripts.c2ctool:main",
            "db2pot = c2cgeoportal.scripts.db2pot:main",
            "themev1tov2 = c2cgeoportal.scripts.themev1tov2:main",
            "theme2fts = c2cgeoportal.scripts.theme2fts:main",
            "l10nv1tov2 = c2cgeoportal.scripts.l10nv1tov2:main",
            "import-ngeo-apps = c2cgeoportal.scripts.import_ngeo_apps:main",
            "gen-version = c2cgeoportal.scripts.gen_version:main",
            "ngeo-version = c2cgeoportal.scripts.import_ngeo_apps:ngeo_git_version",
            "create-demo-theme = c2cgeoportal.scripts.create_demo_theme:main",
            "treeitem_uniquename = c2cgeoportal.scripts.treeitem_uniquename:main",
        ],
        "pyramid.scaffold": [
            "c2cgeoportal_create = c2cgeoportal.scaffolds:TemplateCreate",
            "c2cgeoportal_update = c2cgeoportal.scaffolds:TemplateUpdate",
        ],
        "fanstatic.libraries": [
            "admin = c2cgeoportal.forms:fanstatic_lib",
        ],
        "lingua.extractors": [
            "geomapfish-theme = c2cgeoportal.lib.lingua_extractor:GeoMapfishThemeExtractor",
            "geomapfish-config = c2cgeoportal.lib.lingua_extractor:GeoMapfishConfigExtractor",
            "geomapfish-angular = c2cgeoportal.lib.lingua_extractor:GeoMapfishAngularExtractor",
        ],
    }
)
